from .catalog import SourceCatalog, StationCatalog, ScanCatalog, \
    SourcePlotCatalog
from .sched import parameter
from . import util
from .plot_util import shut_up_mpl

import schedlib as s

# import matplotlib and call .use immediately
import matplotlib
matplotlib.use("Qt5Agg")
matplotlib.rcParams["toolbar"] = "toolmanager"

# in matplotlib version 3.3 the internal epoch was changed from year 1 to 1970
# this epoch is used to convert dates to floating points internally
# unfortunately that internal representation leaks to axis.get_xlim
# so set the epoch to the old version here to get consistent behaviour with
# both version (<3.3 and >=3.3)
from distutils.version import LooseVersion
if LooseVersion(matplotlib.__version__) >= LooseVersion("3.3"):
    matplotlib.rcParams["date.epoch"] = "0000-12-31T00:00:00"

# import function to improve toolbar, but might not be available on all systems
try:
    from .plot_toolbar import adjust as adjust_toolbar
except ImportError:
    def adjust_toolbar(*args, **kwargs):
        pass

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.figure
import matplotlib.pyplot as plt
import matplotlib.dates
import matplotlib.units
import matplotlib.cbook
from matplotlib.colors import to_hex

# hack around bug in formlayout, set an environment variable to force use of Qt5
import os; os.environ["QT_API"] = "pyqt5"
from formlayout import ColorButton

import numpy as np

from astropy.time import Time

from PyQt5.QtWidgets import QApplication, QWidget, QFrame, QScrollArea, \
    QTabWidget, QPushButton, QCheckBox, QGroupBox, QLabel, QLineEdit, \
    QComboBox, QMenu, QRadioButton, \
    QVBoxLayout, QHBoxLayout, QGridLayout, QButtonGroup
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QCursor, QDoubleValidator, QColor

import itertools
from collections import defaultdict, OrderedDict
import math
from datetime import datetime, timedelta, date
from contextlib import contextmanager

# list of colors to cycle through
color_list = [to_hex(c) for c in "bgrcmyk"] + ["orange", "brown"] 

@contextmanager
def wait_cursor():
    QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
    yield
    QApplication.restoreOverrideCursor()

# update matplotlib date converter to handle None, 
# as that value is used to signal a line segment end
def _to_ordinalf(d):
    if d is None:
        return None
    return matplotlib.dates._to_ordinalf(d)
_to_ordinalf_np_vectorized = np.vectorize(_to_ordinalf)
class NoneDateConverter(type(matplotlib.units.registry[datetime])):
    @staticmethod
    def convert(value, unit, axis):
        if matplotlib.units.ConversionInterface.is_numlike(value):
            return value
        else:
            if not np.iterable(value):
                return _to_ordinalf(value)
            else:
                d = np.asarray(value)
                if not d.size:
                    return d
                return _to_ordinalf_np_vectorized(d)
matplotlib.units.registry[datetime] = NoneDateConverter()
time_formatter = matplotlib.dates.DateFormatter("%H:%M:%S")

def mjd2utc(mjd):
    return Time(mjd, format="mjd").datetime

def mjd2lst(mjd, longitude):
    # the base date used is deterministic but undefined
    # (only) useful for plotting
    _, _, lstday = s.sidtim(mjd, longitude, math.pi * 2)
    return epoch2datetime(lstday)

def epoch2datetime(x):
    fractional, integer = math.modf(x)
    return datetime.fromordinal(int(integer)) + \
        timedelta(seconds=round(fractional * 86400))

class FloatEdit(QLineEdit):
    def __init__(self, initial="0.0", *args, **kwargs):
        super().__init__(initial, *args, **kwargs)
        self.setValidator(QDoubleValidator(self))

    def value(self):
        try:
            return float(self.text())
        except ValueError:
            return None

class MarkerSelection(QComboBox):
    description_character = {
        "point": ".",
        "pixel": ",",
        "circle": "o",
        "triangle_down": "v",
        "triangle_up": "^",
        "triangle_left": "<",
        "triangle_right": ">",
        "tri_down": "1",
        "tri_up": "2",
        "tri_left": "3",
        "tri_right": "4",
        "octagon": "8",
        "square": "s",
        "pentagon": "p",
        "star": "*",
        "hexagon1": "h",
        "hexagon2": "H",
        "plus": "+",
        "x": "x",
        "diamond": "D",
        "thin_diamond": "d",
        "vline": "|",
        "hline": "_",
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self.addItems(sorted(self.description_character.keys()))

    def get_marker(self):
        return self.description_character[self.currentText()]

class SourcesWidget(QGroupBox):
    def __init__(self, sources, parent=None):
        super().__init__("Sources", parent)
        main_layout = QVBoxLayout(self)
        scroll = QScrollArea(self)
        main_layout.addWidget(scroll)
        sources_widget = QWidget(self)
        sources_layout = QVBoxLayout(sources_widget)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.source_checkbox = {}
        for source in sources:
            alias = source.aliases[0]
            if len(source.aliases) > 1:
                label = "{} ({})".format(alias, ", ".join(source.aliases[1:]))
            else:
                label = alias
            checkbox = QCheckBox(label, sources_widget)
            checkbox.setChecked(True)
            sources_layout.addWidget(checkbox)
            self.source_checkbox[alias] = checkbox
        all_button = QPushButton("All", self)
        scroll.setWidget(sources_widget)
        main_layout.addWidget(all_button)
        all_button.clicked.connect(self.change_all)

    def selected_sources(self):
        return [source for source, checkbox in self.source_checkbox.items()
                if checkbox.isChecked()]

    def change_all(self):
        to = not any(checkbox.isChecked() 
                     for checkbox in self.source_checkbox.values())
        for checkbox in self.source_checkbox.values():
            checkbox.setChecked(to)

class PlotCheckBox(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        
        self.check_box = QCheckBox("", self)
        self.clicked = self.check_box.clicked
        layout.addWidget(self.check_box)

        self.figure = matplotlib.figure.Figure()
        canvas = FigureCanvas(self.figure)
        layout.addWidget(canvas)
        axis = self.figure.add_subplot(111)
        axis.xaxis.set_visible(False)
        axis.yaxis.set_visible(False)
        self.line = axis.plot([0,1], [0,1], "r-")[0]
        # force the plot to have about the same size as the check box
        # for some reason the check box size method return a very high 
        # number for width, so use height for both dimensions
        size = self.check_box.size()
        canvas.setFixedSize(size.height(), size.height())
        canvas.draw()

    def isChecked(self):
        return self.check_box.isChecked()

    def setChecked(self, to):
        return self.check_box.setChecked(to)

    def get_color(self):
        return self.line.get_color()

    def set_color(self, color):
        r = self.line.set_color(color)
        self.figure.canvas.draw()
        return r

    def get_visible(self):
        return self.figure.get_visible()

    def set_visible(self, to):
        r = self.figure.set_visible(to)
        self.figure.canvas.draw()
        return r

    def get_linewidth(self):
        return self.line.get_linewidth()

    def set_linewidth(self, width):
        r = self.line.set_linewidth(width)
        self.figure.canvas.draw()
        return r

    def get_properties(self):
        return plt.getp(self.line)

class HighlightGroup(QWidget):
    choices = ("Hide", "Show", "Mark")
    choices_text = {c: c[0] for c in choices}
    text_choices = {c[0]: c for c in choices}
    clicked = pyqtSignal()
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        self.buttons = QButtonGroup(self)
        for index, choice in enumerate(self.choices):
            button = QRadioButton(self.choices_text[choice])
            button.clicked.connect(self.clicked)
            self.buttons.addButton(button, index)
            layout.addWidget(button)

    def current_text(self):
        return self.text_choices[self.buttons.checkedButton().text()]

    def select(self, text):
        self.buttons.button(self.choices.index(text)).setChecked(True)
        
    def setChecked(self, to):
        if to:
            self.select("Show")
        else:
            self.select("Hide")

class BaselinesWidget(QGroupBox):
    changed = pyqtSignal()
    def _toggle(self, widgets):
        check = not any(w.isChecked() for w in widgets)
        for w in widgets:
            w.setChecked(check)
        self.changed.emit()
    
    def _handle_row(self, station):
        self._toggle([box for (row, _), box in self.check_box.items()
                      if row == station])

    def _handle_column(self, station):
        self._toggle([box for (_, column), box in self.check_box.items()
                      if column == station])

    def _handle_station(self, station):
        self._toggle([box for baseline, box in self.check_box.items()
                      if station in baseline])

    def _handle_all(self):
        self._toggle(list(self.check_box.values()))

    def __init__(self, stations, title, checkbox_type, parent=None):
        super().__init__(title, parent)
        scroll_layout = QVBoxLayout(self)
        self.setLayout(scroll_layout)
        scroll_area = QScrollArea(self)
        scroll_layout.addWidget(scroll_area)
        scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll_widget = QWidget(scroll_area)
        selection_layout = QGridLayout(scroll_widget)
        self.check_box = {} # {(station_row, station_column): checkbox_type()}
        all_button = QPushButton("All", scroll_widget)
        all_button.clicked.connect(self._handle_all)
        selection_layout.addWidget(all_button, 0, 0)

        for column, header in enumerate(stations[1:], 1):
            button = QPushButton("\n".join(header.station), scroll_widget)
            button.clicked.connect(lambda _, s=header.station: 
                                   self._handle_station(s))
            selection_layout.addWidget(button, 0, column)

        for row, header in enumerate(stations[:-1], 1):
            button = QPushButton(header.station, scroll_widget)
            button.clicked.connect(lambda _, s=header.station: 
                                   self._handle_station(s))
            selection_layout.addWidget(button, row, 0)
            for column, item in enumerate(stations[row:], row):
                check_box = checkbox_type(parent=scroll_widget)
                check_box.setChecked(True)
                check_box.clicked.connect(self.changed)
                self.check_box[(header.station, item.station)] = check_box
                selection_layout.addWidget(check_box, row, column)
        scroll_area.setWidget(scroll_widget)

class BaselineSelectionWidget(BaselinesWidget):
    def get_visible(self, baseline):
        return self.check_box[baseline].isChecked()

class BaselineHighlightWidget(BaselinesWidget):
    def __init__(self, stations, parent=None):
        super().__init__(stations, 
                         "Select H(ide), S(how) or M(ark) per baseline", 
                         HighlightGroup, 
                         parent)

    def _toggle(self, widgets):
        min_index = min(HighlightGroup.choices.index(w.current_text()) 
                        for w in widgets)
        to = HighlightGroup.choices[
            (min_index + 1) % len(HighlightGroup.choices)]
        for w in widgets:
            w.select(to)

    def get_properties(self, baseline):
        highlight = self.check_box[baseline].current_text() == "Mark"
        return {"linewidth": 3 if highlight else 1,
                "color": "b" if highlight else "r"}

    def get_visible(self, baseline):
        return self.check_box[baseline].current_text() != "Hide"


class BaselineConfigurationWidget(QWidget):
    def _set_plot(self):
        to = self.plot_check_box.isChecked()
        for w in self.baselines.check_box.values():
            if w.isChecked():
                w.set_visible(to)

    def _set_color(self):
        color = self.color_button.get_color().name()
        for w in self.baselines.check_box.values():
            if w.isChecked():
                w.set_color(color)

    def _set_width(self):
        width = self.width_edit.value()
        if width is not None:
            for w in self.baselines.check_box.values():
                if w.isChecked():
                    w.set_linewidth(width)

    def check_update(self):
        if self.automatic_update.isChecked():
            attributes = {"color": (lambda c: self.color_button.set_color(
                              QColor(c))), 
                          "visible": self.plot_check_box.setChecked, 
                          "linewidth": (lambda v: self.width_edit.setText(
                              str(v)))}
            value_count = {attribute: defaultdict(int) 
                           for attribute in attributes.keys()}
            for w in self.baselines.check_box.values():
                if w.isChecked():
                    for attribute in attributes.keys():
                        value = getattr(w, "get_" + attribute)()
                        value_count[attribute][value] += 1
            if len(value_count["color"]) == 0:
                return # nothing selected, do nothing
            for attribute, function in attributes.items():
                value = max((count, value) for value, count in 
                            value_count[attribute].items())[1]
                function(value)

    def get_properties(self, baseline):
        check_box = self.baselines.check_box[baseline]
        return {"linewidth": check_box.get_linewidth(),
                "color": check_box.get_color()}

    def get_visible(self, baseline):
        return self.baselines.check_box[baseline].get_visible()

    def __init__(self, stations, parent=None):
        super().__init__(parent)
        master_layout = QVBoxLayout(self)
        
        self.baselines = BaselinesWidget(
            stations, "Select baselines to configure", PlotCheckBox, self)
        self.baselines.changed.connect(self.check_update)
        master_layout.addWidget(self.baselines)

        group_box = QGroupBox("Configure selection", self)
        control_layout = QHBoxLayout(group_box)
        control_layout.addStretch(1)
        def add_separator():
            control_layout.addStretch(1)
            line = QFrame(group_box)
            line.setFrameShape(QFrame.VLine)
            line.setFrameShadow(QFrame.Sunken)
            line.setFixedWidth(3)
            control_layout.addWidget(line)
            control_layout.addStretch(1)

        self.automatic_update = QCheckBox("Update", group_box)
        self.automatic_update.setToolTip(
            "If checked, configuration values are updated when selection "
            "changes, to the most common values")
        self.automatic_update.setChecked(True)
        control_layout.addWidget(self.automatic_update)
        add_separator()
        
        self.plot_check_box = QCheckBox("Plot", group_box)
        control_layout.addWidget(self.plot_check_box)
        button = QPushButton("Set", group_box)
        button.clicked.connect(self._set_plot)
        control_layout.addWidget(button)
        add_separator()
        
        control_layout.addWidget(QLabel("Color", group_box))
        self.color_button = ColorButton(group_box)
        control_layout.addWidget(self.color_button)
        button = QPushButton("Set", group_box)
        button.clicked.connect(self._set_color)
        control_layout.addWidget(button)
        add_separator()

        control_layout.addWidget(QLabel("Width", group_box))
        self.width_edit = FloatEdit("1.0", group_box)
        control_layout.addWidget(self.width_edit)
        button = QPushButton("Set", group_box)
        button.clicked.connect(self._set_width)
        control_layout.addWidget(button)
        
        control_layout.addStretch(1)
        master_layout.addWidget(group_box)

        self.check_update()

class StationsWidget(QGroupBox):
    def get_visible(self, station):
        return self.visible_checkbox[station].isChecked()

    def get_color(self, station):
        return self.color_button[station].get_color().name()

    def get_linewidth(self, station):
        return self.linewidth_edit[station].value()

    def get_properties(self, station):
        ret = {"color": self.get_color(station)}
        lw = self.get_linewidth(station)
        if lw is not None:
            ret["linewidth"] = lw
        return ret

    def __init__(self, stations, parent=None):
        super().__init__("Stations", parent)
        layout = QGridLayout(self)
        self.visible_checkbox = {}
        self.color_button = {}
        self.linewidth_edit = {}
        for column, label in enumerate(("Plot", "Color", "Linewidth"), 1):
            layout.addWidget(QLabel(label, self), 0, column)
        color_cycle = itertools.cycle(color_list)
        for row, station in enumerate(stations, 1):
            name = station.station
            layout.addWidget(QLabel(name, self), row, 0)
            
            checkbox = QCheckBox(self)
            checkbox.setChecked(True)
            self.visible_checkbox[name] = checkbox
            layout.addWidget(checkbox, row, 1)

            color = ColorButton(self)
            color.set_color(QColor(next(color_cycle)))
            self.color_button[name] = color
            layout.addWidget(color, row, 2)

            width = FloatEdit("1.0", self)
            self.linewidth_edit[name] = width
            layout.addWidget(width, row, 3)
            
class UVWidget(QWidget):
    def _check_axis(self):
        self.axis_unit_multiplier.setVisible(
            self.axis_unit.currentText() == "Wv")
            
    wavelength_multiplier = OrderedDict((("λ", 1),
                                         ("Kλ", 1e-3),
                                         ("Mλ", 1e-6)))

    def get_unit_multiplier(self):
        if self.axis_unit.currentText() == "Wv":
            unit = self.axis_unit_multiplier.currentText()
            return (unit, self.wavelength_multiplier[unit])
        else:
            return ("Km", 1)

    def __init__(self, sources, stations, parent=None):
        super().__init__(parent)
        master_layout = QVBoxLayout(self)
        
        top_layout = QHBoxLayout()
        top_layout.addWidget(QLabel("Axis units", self))
        self.axis_unit = QComboBox(self)
        self.axis_unit.addItems(["Km", "Wv"])
        self.axis_unit.currentTextChanged.connect(self._check_axis)
        top_layout.addWidget(self.axis_unit)
        self.axis_unit_multiplier = QComboBox(self)
        self.axis_unit_multiplier.addItems(self.wavelength_multiplier.keys())
        top_layout.addWidget(self.axis_unit_multiplier)
        top_layout.addStretch(1)
        self._check_axis()
        self.sources = SourcesWidget(sources, self)
        top_layout.addWidget(self.sources)
        top_layout.setStretchFactor(self.sources, 1)
        master_layout.addLayout(top_layout)

        self.tab_widget = QTabWidget(self)
        self.tab_widget.addTab(
            BaselineHighlightWidget(stations, self.tab_widget), "Highlight")
        self.tab_widget.addTab(
            BaselineConfigurationWidget(stations, self.tab_widget), "Advanced")
        master_layout.addWidget(self.tab_widget)

    def get_configuration_widget(self):
        return self.tab_widget.currentWidget()

class XYBaseWidget(QWidget): # shared by XY and Uptime
    def _check_x_axis(self):
        x = self.x_axis.currentText()
        self.ut_offset.setVisible(x == "UT")
        self.lst_base.setVisible(x == "LST")

    def get_x_axis(self):
        return self.x_axis.currentText()

    def get_ut_offset(self):
        return int(self.ut_offset.currentText())

    def get_lst_base(self):
        return self.lst_base.currentText()

    def get_y_axis(self):
        if len(self.y_items) > 1:
            return self.y_axis.currentText()
        else:
            return self.y_items[0]
    
    def __init__(self, x_items, y_items, sources, stations, parent=None):
        super().__init__(parent)
        self.y_items = y_items
        master_layout = QHBoxLayout(self)
        
        left_layout = QVBoxLayout()
        left_layout.addStretch(1)

        top_left_layout = QHBoxLayout()
        left_layout.addStretch(1)
        top_left_layout.addWidget(QLabel("X", self))
        self.x_axis = QComboBox(self)
        self.x_axis.addItems(x_items)
        self.x_axis.currentTextChanged.connect(self._check_x_axis)
        top_left_layout.addWidget(self.x_axis)
        self.ut_offset = QComboBox(self)
        self.ut_offset.addItems(str(i) for i in range(-11, 13))
        self.ut_offset.setCurrentText("0")
        top_left_layout.addWidget(self.ut_offset)
        self.lst_base = QComboBox(self)
        self.lst_base.addItems(["Greenwich"] + [e.station for e in stations])
        top_left_layout.addWidget(self.lst_base)
        top_left_layout.addStretch(1)
        if len(self.y_items) > 1:
            top_left_layout.addWidget(QLabel("Y", self))
            self.y_axis = QComboBox(self)
            self.y_axis.addItems(y_items)
            top_left_layout.addWidget(self.y_axis)

        self._check_x_axis()
        left_layout.addLayout(top_left_layout)
        
        left_layout.addStretch(1)
        
        self.stations = StationsWidget(stations, self)
        left_layout.addWidget(self.stations)
        left_layout.addStretch(1)

        master_layout.addLayout(left_layout)

        self.sources = SourcesWidget(sources, self)
        master_layout.addWidget(self.sources)

class XYWidget(XYBaseWidget):
    def __init__(self, sources, stations, parent=None):
        y_items = ["AZ", "EL",  "HA", "PA", "Sec"]
        x_items = ["UT", "GST", "LST"] + y_items
        super().__init__(x_items, y_items, sources, stations, parent)

class UptimeWidget(XYBaseWidget):
    def __init__(self, sources, stations, parent=None):
        x_items = ["UT", "GST", "LST"]
        y_items = ["Antenna"]
        super().__init__(x_items, y_items, sources, stations, parent)

class RADecWidget(QWidget):
    def get_visible(self, calcode):
        return self.widgets[calcode]["Plot"].isChecked()

    def get_properties(self, calcode):
        widgets = self.widgets[calcode]
        ret = {"marker": widgets["Marker"].get_marker(),
               "color": widgets["Color"].get_color().name()}
        lw = widgets["Linewidth"].value()
        if lw is not None:
            ret["linewidth"] = lw
        s = widgets["Size"].value()
        if s is not None:
            ret["s"] = s
        return ret

    calcode_map = {
        "Y": "VLA",
        "N": "USNO",
        "M": "MERLIN",
        "V": "VLBA"
    }
    marker_map = {
        "Y": "square",
        "N": "triangle_right",
        "M": "thin_diamond",
        "V": "star"
    }
    def __init__(self, sources, parent=None):
        super().__init__(parent)
        calcodes = set(s.srlcalc for s in sources)
        
        layout = QGridLayout(self)

        columns = OrderedDict([
            ("Plot", QCheckBox),
            ("Marker", MarkerSelection),
            ("Color", ColorButton), 
            ("Size", lambda parent: FloatEdit("100", parent)), 
            ("Linewidth", lambda parent: FloatEdit("3", parent))])

        self.widgets = {}

        layout.addWidget(QLabel("Catalog", self), 0, 0, Qt.AlignBottom)
        for column, label in enumerate(columns.keys(), 1):
            layout.addWidget(QLabel(label, self), 0, column, Qt.AlignBottom)

        color_cycle = itertools.cycle(color_list)
        marker_cycle = itertools.cycle(sorted(
            set(MarkerSelection.description_character.keys()) -
            set(self.marker_map.values())))
        for row, calcode in enumerate(sorted(calcodes) + ["SCHED"], 1):
            catalog = self.calcode_map.get(calcode, calcode)
            layout.addWidget(QLabel(catalog, self), row, 0)
            widgets = {}
            for column, (column_label, widget_function) in \
                enumerate(columns.items(), 1):
                w = widget_function(self)
                widgets[column_label] = w
                layout.addWidget(w, row, column)

            self.widgets[calcode] = widgets
            # initialize widgets
            widgets["Plot"].setChecked(True)
            if calcode in self.marker_map:
                marker = self.marker_map[calcode]
            else:
                marker = next(marker_cycle)
            widgets["Marker"].setCurrentText(marker)
            widgets["Color"].set_color(QColor(next(color_cycle)))

class BeamWidget(QWidget):
    def get_source(self):
        return self.source.currentText()

    def get_setup_index(self):
        return self.wave_index[self.wavelength.currentIndex()][1]

    def get_oversampling(self):
        return int(self.oversampling.currentText())

    def get_weighting(self):
        return self.weight.currentText()

    def get_image_transfer_function(self):
        return self.image_transfer_function.currentText()

    def get_color_map(self):
        return matplotlib.cm.get_cmap(self.color_map.currentText())

    def get_contour_colors(self):
        return matplotlib.cm.get_cmap(self.contour_colors.currentText())

    def get_contours(self):
        return self.contours.isChecked()

    def __init__(self, sources, stations, scans, parent=None):
        super().__init__(parent)
        master_layout = QVBoxLayout(self)
        top_layout = QGridLayout()
        
        used_setups = set(scan.setnum for scan in scans)
        self.wave_index = sorted((300 / s.schsf.sffreq[0, i-1], i)
                                 for i in used_setups)
        for row, (label, attribute, options) in enumerate((
                ("Source", "source", [s.aliases[0] for s in sources]),
                ("Wavelength (cm)", "wavelength", 
                 ["{:.1f}".format(w[0] * 100).rstrip("0").rstrip(".") + 
                  " ({})".format(util.f2str(s.schssf.setfile[w[1]]))
                  for w in self.wave_index]),
                ("Oversampling factor", "oversampling", map(str, range(1, 11))),
                ("Weight", "weight", ["Natural", "Uniform"]),
                ("Image transfer function", "image_transfer_function", 
                 ["Linear", "Logarithmic", "Square-root"]),
                ("Color palette", "color_map", 
                 sorted(plt.colormaps(), key=lambda x: x.lower()))
        )):
            top_layout.addWidget(QLabel(label, self), row, 0)
            w = QComboBox(self)
            setattr(self, attribute, w)
            w.addItems(options)
            top_layout.addWidget(w, row, 1)
        self.color_map.setCurrentText("Greys")
        
        self.contours = QCheckBox("Plot contours", self)
        top_layout.addWidget(self.contours, row + 1, 0)
        self.contour_colors = QComboBox(self)
        self.contour_colors.addItems(sorted(plt.colormaps(), 
                                            key=lambda x: x.lower()))
        self.contour_colors.setCurrentText("Spectral")
        top_layout.addWidget(self.contour_colors, row + 1, 1)

        master_layout.addLayout(top_layout)
        
        self.baselines = BaselineSelectionWidget(
            stations, "Baselines to include", QCheckBox, self)
        master_layout.addWidget(self.baselines)

class MainWidget(QWidget):
    finished = pyqtSignal()
    
    def show_plot(self):
        active = self.tab_widget.tabText(self.tab_widget.currentIndex())
        {"UV": self.plot_uv,
         "XY": self.plot_xy,
         "Uptime": self.plot_uptime,
         "RA-Dec": self.plot_radec,
         "Beam": self.plot_beam}[active]()

    def __init__(self, is_restart, parent=None):
        super().__init__(parent)
        window_layout = QVBoxLayout(self)
        self.setLayout(window_layout)
        
        source_catalog = SourceCatalog()
        source_catalog.read()
        source_catalog.set_aliases()
        self.sources = source_catalog.used()
        self.radec_sources = [e for e in SourcePlotCatalog().read() 
                              if not e.srlused]
        self.scan_catalog = ScanCatalog()
        self.scan_catalog.read()
        station_catalog = StationCatalog()
        station_catalog.read()
        station_catalog.read_scheduled_attributes()
        self.stations = station_catalog.used()
        
        self.tab_widget = QTabWidget(self)
        self.uv = UVWidget(self.sources, self.stations, self.tab_widget)
        self.tab_widget.addTab(self.uv, "UV")
        self.xy = XYWidget(self.sources, self.stations, self.tab_widget)
        self.tab_widget.addTab(self.xy, "XY")
        self.uptime = UptimeWidget(self.sources, self.stations, self.tab_widget)
        self.tab_widget.addTab(self.uptime, "Uptime")
        self.radec = RADecWidget(self.radec_sources, self.tab_widget)
        self.tab_widget.addTab(self.radec, "RA-Dec")
        self.beam = BeamWidget(self.sources, self.stations, 
                               self.scan_catalog.used(), self.tab_widget)
        self.tab_widget.addTab(self.beam, "Beam")
        
        window_layout.addWidget(self.tab_widget)

        control_widget = QWidget(self)
        control_layout = QHBoxLayout(control_widget)
        control_widget.setLayout(control_layout)
        plot_button = QPushButton("Plot", control_widget)
        control_layout.addWidget(plot_button)
        plot_menu = QMenu("Plot", plot_button)
        plot_button.clicked.connect(self.show_plot)
        restart_button = QPushButton("Restart", control_widget)
        control_layout.addWidget(restart_button)
        if not is_restart:
            finish_button = QPushButton("Finish", control_widget)
            control_layout.addWidget(finish_button)
            finish_button.clicked.connect(lambda: self.done(1))
        exit_button = QPushButton("Exit", control_widget)
        control_layout.addWidget(exit_button)
        exit_button.clicked.connect(lambda: self.done(0))
        restart_button.clicked.connect(lambda: self.done(2))
        window_layout.addWidget(control_widget)

        self._result = None

    # replicate signal functionality of a QDialog, without the window behaviour
    def done(self, r):
        self._result = r
        self.finished.emit()

    def result(self):
        # 0 -> exit, 1 -> finish, 2 -> restart
        # just exit if no result has been set (window closed manually)
        return self._result if self._result is not None else 0

    def plot_uv(self):
        with wait_cursor():
            plot_sources = self.uv.sources.selected_sources()
            if len(plot_sources) == 0:
                return
            alias_source = {}
            for source in self.sources:
                if source.aliases[0] in plot_sources:
                    for alias in source.aliases:
                        alias_source[alias] = source.aliases[0]
            
            configuration = self.uv.get_configuration_widget()
            # plot sources in a square pattern, expanding columns before rows
            items = len(plot_sources)
            columns = math.ceil(math.sqrt(items))
            rows = math.ceil(items / columns)
            with shut_up_mpl():
                figure, axes = plt.subplots(rows, columns, squeeze=False)
            figure.canvas.set_window_title("UV Plot")
            axes = axes.reshape((rows * columns,))

            unit, multiplier = self.uv.get_unit_multiplier()
            scan_multiplier = multiplier
            
            # Get parameters for plotting of multiple lines for MFS.
            # MFSRAT is the ratio of the upper to the lower frequency.
            # NMFS is the number of frequencies to use.
            if s.schcon.nmfs > 1:
                # multi frequency synthesis plots requested
                mfs_base = 2 / (1 + s.schcon.mfsrat)
                mfs_inc = mfs_base * (s.schcon.mfsrat - 1) / (s.schcon.nmfs - 1)
            else:
                mfs_base = 1
                mfs_inc = 0
            
            source_baseline_xy = defaultdict(
                lambda: defaultdict(lambda: [[], []]))
            for scan_index, scan in enumerate(self.scan_catalog.used(), 
                                              self.scan_catalog.scan_offset):
                if scan.scnsrc not in alias_source:
                    continue
                source = alias_source[scan.scnsrc]
                
                if unit.endswith("λ"):
                    frequency = s.schsf.sffreq[0, scan.setnum-1]
                    # frequency is in MHz
                    wavelength = 3e8 / (frequency * 1e6)
                    # data is in Km
                    scan_multiplier = multiplier / wavelength * 1000

                station_uv = {station: np.array(s.stauv(scan_index+1, i+1)).\
                              reshape((2, 2)) * scan_multiplier
                              for i, station in enumerate(self.stations)}
                for mfs in (mfs_base + i * mfs_inc 
                            for i in range(s.schcon.nmfs)):
                    mfs_station_uv = {k: v * mfs for k, v in station_uv.items()}
                
                    baselines = itertools.combinations(self.stations, 2)
                    for station1, station2 in baselines:
                        station_names = (station1.station, station2.station)
                        if configuration.get_visible(station_names) and \
                           station1.stascn[scan_index] and \
                           station2.stascn[scan_index]:
                            diff = mfs_station_uv[station1] - \
                                   mfs_station_uv[station2]
                            xy = source_baseline_xy[source][station_names]
                            for i in range(2): # loop over x (u) and y (v)
                                l = xy[i]
                                l.extend(diff[i])
                                l.append(None)
                                l.extend(-diff[i])
                                l.append(None)
            for source, baseline_xy in source_baseline_xy.items():
                source_index = plot_sources.index(source)
                axis = axes[source_index]
                axis.set_aspect("equal")
                axis.set_title(source)
                axis.xaxis.set_label_text("U ({})".format(unit))
                axis.yaxis.set_label_text("V ({})".format(unit))
                for baseline, [x, y] in baseline_xy.items():
                        axis.plot(
                            x, y, label="{} - {}".format(*baseline),
                            **configuration.get_properties(baseline))
            figure.subplots_adjust(left=0.05, right=0.95,
                                   bottom=0.05, top=0.95,
                                   wspace=.5, hspace=.5)
            # hide unused axes
            for axis in axes[len(plot_sources):]:
                axis.set_visible(False)

            adjust_toolbar(figure)
            figure.tight_layout()

    def plot_xy(self):
        with wait_cursor():
            plot_sources = self.xy.sources.selected_sources()
            if len(plot_sources) == 0:
                return
            alias_source = set()
            for source in self.sources:
                if source.aliases[0] in plot_sources:
                    alias_source.update(source.aliases)

            with shut_up_mpl():
                figure, axis = plt.subplots()
            station_xy = defaultdict(lambda: [[], []])

            axis_type = [self.xy.get_x_axis(), 
                         self.xy.get_y_axis()]

            figure.canvas.set_window_title("{}-{} Plot".format(
                *axis_type[::-1]))

            axis_function = []
            for type_ in axis_type:
                if type_ == "UT":
                    td = timedelta(hours=self.xy.get_ut_offset())
                    def f(scan_index, station, td=td):
                        scan = self.scan_catalog.entries[scan_index]
                        return [mjd2utc(scan.startj) + td, 
                                mjd2utc(scan.stopj) + td,
                                None]
                elif type_ in ["GST", "LST"]:
                    if type_ == "LST":
                        lst_base = self.xy.get_lst_base()
                        if lst_base == "Greenwich":
                            longitude = 0
                        else:
                            longitude = next(station.long_bn 
                                             for station in self.stations
                                             if station.station == lst_base)
                    else:
                        longitude = 0
                    def f(scan_index, station, longitude=longitude): 
                        start = mjd2lst(scan.startj, longitude)
                        stop = mjd2lst(scan.stopj, longitude)
                        return [start, stop, None]
                elif type_ == "AZ":
                    def f(scan_index, station):
                        return [station.az1[scan_index] % 360,
                                station.az2[scan_index] % 360,
                                None]
                elif type_ in ["EL", "HA", "PA"]:
                    start_attribute = "{}{}".format(type_.lower(), 1)
                    end_attribute = "{}{}".format(type_.lower(), 2)
                    def f(scan_index, station, 
                          sa=start_attribute, ea=end_attribute):
                        return [getattr(station, sa)[scan_index],
                                getattr(station, ea)[scan_index],
                                None]
                elif type_ == "Sec":
                    def f(scan_index, station):
                        start = 1 / math.sin(
                            station.el1[scan_index] * parameter.raddeg)
                        end = 1 / math.sin(
                            station.el2[scan_index] * parameter.raddeg)
                        return [start, end, None]
                else:
                    raise RuntimeError("Unhandled axis type {}".format(type_))
                axis_function.append(f)


            for scan_index, scan in enumerate(self.scan_catalog.used(), 
                                              self.scan_catalog.scan_offset):
                if scan.scnsrc not in alias_source:
                    continue
                for station in self.stations:
                    if not (self.xy.stations.get_visible(station.station) and
                            station.stascn[scan_index]):
                        continue

                    xy = station_xy[station]
                    for axis_index, function in enumerate(axis_function):
                        xy[axis_index].extend(function(scan_index, station))
            
            axis.set_title("{} vs {}".format(*axis_type[::-1]))
            axis.xaxis.set_label_text(axis_type[0])
            axis.yaxis.set_label_text(axis_type[1])
            for station, [x, y] in station_xy.items():
                name = station.station
                axis.plot(x, y, label=name,
                          **self.xy.stations.get_properties(name))

            if axis_type[0] in ["UT", "LST", "GST"]:
                axis.xaxis.set_major_formatter(time_formatter)
                figure.autofmt_xdate()
                if axis_type[0] == "UT":
                    # Cludge to update the label of a time axis to reflect the
                    # day of the displayed times.
                    # Updating the axis label within the draw event doesn't 
                    # update the draw title, so have to call draw again. 
                    # Use a boolean to prevent infinite recursion.
                    call_draw = True
                    def draw_event(event):
                        nonlocal call_draw
                        if call_draw:
                            start = epoch2datetime(axis.get_xlim()[0])
                            axis.xaxis.set_label_text(
                                f"Starting at UT {str(start)}")
                            call_draw = False
                            figure.canvas.draw()
                        else:
                            call_draw = True
                    figure.canvas.mpl_connect("draw_event", draw_event)
                else:
                    axis.xaxis.lst_base_date = date.fromordinal(
                        int(axis.get_xlim()[0]))
            if axis_type[0] == "Sec":
                axis.invert_xaxis()
            if axis_type[1] == "Sec":
                axis.invert_yaxis()
            
            legend = axis.legend()
            legend.set_draggable(True)
            adjust_toolbar(figure, axis_type[0])
            figure.tight_layout()

    def plot_uptime(self):
        with wait_cursor():
            plot_sources = self.uptime.sources.selected_sources()
            if len(plot_sources) == 0:
                return
            alias_source = {}
            for source in self.sources:
                if source.aliases[0] in plot_sources:
                    for alias in source.aliases:
                        alias_source[alias] = source.aliases[0]

            stations = [station for station in self.stations
                        if self.uptime.stations.get_visible(station.station)]
            
            with shut_up_mpl():
                figure, axes = plt.subplots(len(plot_sources), squeeze=False)
            axes = axes.reshape((len(plot_sources),))

            figure.canvas.set_window_title("Uptime Plot")

            axis_function = []
            axis_type = self.uptime.get_x_axis()
            if axis_type == "UT":
                td = timedelta(hours=self.uptime.get_ut_offset())
                def time_function(mjd, td=td):
                    return mjd2utc(mjd) + td 
            elif axis_type in ["GST", "LST"]:
                if axis_type == "LST":
                    lst_base = self.uptime.get_lst_base()
                    if lst_base == "Greenwich":
                        longitude = 0
                    else:
                        longitude = next(station.long_bn 
                                         for station in self.stations
                                         if station.station == lst_base)
                else:
                    longitude = 0
                def time_function(mjd, longitude=longitude): 
                    return mjd2lst(mjd, longitude)
            else:
                raise RuntimeError("Unhandled axis type {}".format(axis_type))

            source_station_left = defaultdict(lambda: defaultdict(list))
            source_station_width = defaultdict(lambda: defaultdict(list))
            scans = self.scan_catalog.used()
            for scan_index, scan in enumerate(scans, 
                                              self.scan_catalog.scan_offset):
                if scan.scnsrc not in alias_source:
                    continue
                
                left = time_function(scan.startj)
                width = scan.stopj - scan.startj

                source = alias_source[scan.scnsrc]
                station_left = source_station_left[source]
                station_width = source_station_width[source]
                for station in stations:
                    if station.stascn[scan_index] and \
                       (station.up1[scan_index] == "") and \
                       (station.up2[scan_index] == ""):
                        station_left[station].append(left)
                        station_width[station].append(width)

            legend_line = {}
            for source, station_left in source_station_left.items():
                axis = axes[plot_sources.index(source)]
                for station_index, station in enumerate(stations):
                    if station in station_left:
                        left = station_left[station]
                        bottom = [len(stations) - station_index] * len(left)
                        legend_line[station.station] = axis.barh(
                            y=bottom, left=left,
                            width=source_station_width[source][station],
                            label=station.station,
                            **self.uptime.stations.get_properties(
                                station.station))
            
            start = time_function(scans[0].startj)
            end = time_function(scans[-1].stopj)
            for axis, source in zip(axes, plot_sources):
                axis.xaxis.set_major_formatter(time_formatter)
                axis.xaxis.set_major_locator(matplotlib.dates.AutoDateLocator())
                axis.set_xbound(lower=start, upper=end)
                axis.yaxis.set_label_text(source)
                axis.tick_params(axis="y", which="both", 
                                 left=False, labelleft=False)
            figure.autofmt_xdate()
            if axis_type == "UT":
                # Cludge to update the label of a time axis to reflect the
                # day of the displayed times.
                # Updating the axis label within the draw event doesn't 
                # update the draw title, so have to call draw again. 
                # Use a boolean to prevent infinite recursion.
                call_draw = True
                def draw_event(event):
                    nonlocal call_draw
                    if call_draw:
                        start = epoch2datetime(axis.get_xlim()[0])
                        axis.xaxis.set_label_text(
                            f"Starting at UT {str(start)}")
                        call_draw = False
                        figure.canvas.draw()
                    else:
                        call_draw = True
                figure.canvas.mpl_connect("draw_event", draw_event)
            else:
                axis.xaxis.set_label_text(axis_type)
                for axis in axes:
                    axis.xaxis.lst_base_date = date.fromordinal(
                        int(axis.get_xlim()[0]))
            
            plotted_stations = [station.station for station in stations
                                if station.station in legend_line]
            station_lines = [legend_line[station] 
                             for station in plotted_stations]
            legend = axes[0].legend(station_lines, plotted_stations, ncol=4,
                                    loc="lower left", bbox_to_anchor=(-0.05, 1))
            legend.set_draggable(True)
            adjust_toolbar(figure, axis_type)
            # force a draw event to have tight layout work with
            # how the figure is formatted when displayed
            figure.canvas.draw()
            figure.tight_layout()

    def plot_radec(self):
        with wait_cursor():
            with shut_up_mpl():
                figure, axis = plt.subplots()
            figure.canvas.set_window_title("RA-Dec Plot")
            label_points = {}
            label_annotations = {}
            # catalogs
            calcodes = set(s.srlcalc for s in self.radec_sources)
            for calcode in sorted(calcodes):
                catalog_label = self.radec.calcode_map.get(calcode, calcode)
                if catalog_label == "":
                    catalog_label = "No label"
                catalog_sources = [s for s in self.radec_sources
                                   if s.srlcalc == calcode]
                xy = np.array(
                    [[s.srlra, s.srldec] for s in catalog_sources])
                xy[:, 0] /= parameter.radhr
                xy[:, 1] /= parameter.raddeg
                points = axis.scatter(xy[:, 0], xy[:, 1], label=catalog_label,
                                      **self.radec.get_properties(calcode))
                points.set_visible(self.radec.get_visible(calcode))
                label_points[catalog_label] = points
                
                # plot source names if they dont clutter (and slow down)
                # the plot
                visible = (len(catalog_sources) < 20) and points.get_visible()
                annotations = []
                for i, s in enumerate(catalog_sources):
                    a = axis.annotate(s.srlname, xy=xy[i])
                    a.set_visible(visible)
                    annotations.append(a)
                label_annotations[catalog_label] = annotations


            # sched
            if self.radec.get_visible("SCHED"):
                xy = np.array([[s.ra2000, s.d2000] for s in self.sources])
                xy[:, 0] /= parameter.radhr
                xy[:, 1] /= parameter.raddeg
                label_points["SCHED"] = axis.scatter(
                    xy[:, 0], xy[:, 1], label="SCHED",
                    **self.radec.get_properties("SCHED"))
                annotations = []
                for i, s in enumerate(self.sources):
                    annotations.append(
                        axis.annotate(s.aliases[0], xy=xy[i]))
                    label_annotations["SCHED"] = annotations

            axis.set_title("Dec - RA\n"
                           "Click in the legend to show/hide points/labels")
            axis.xaxis.set_label_text("RA (Hours)")
            axis.yaxis.set_label_text("Dec (Deg)")
            axis.set_aspect(parameter.raddeg / parameter.radhr)
            axis.set_xbound(lower=0, upper=24)
            axis.invert_xaxis()
            axis.set_ybound(lower=-90, upper=90)
            legend = axis.legend()
            visible_alpha = {True: 1, False: 0.2}
            for artist in legend.legendHandles:
                artist.set_picker(5)
                plot_artist = label_points[artist.get_label()]
                artist.set_alpha(visible_alpha[plot_artist.get_visible()])
            for text in legend.get_texts():
                text.set_picker(5)
                annotations = label_annotations[text.get_text()]
                visible = annotations and annotations[0].get_visible()
                text.set_alpha(visible_alpha[visible])
            legend.set_draggable(True)
            # allow clicking in the legend to show/hide points/labels
            def onpick(event):
                artist_type = type(event.artist)
                if artist_type is matplotlib.collections.PathCollection:
                    # scatter plot type
                    plot_artist = label_points[event.artist.get_label()]
                    visible = not plot_artist.get_visible()
                    plot_artist.set_visible(visible)
                    event.artist.set_alpha(visible_alpha[visible])
                    figure.canvas.draw()
                elif artist_type is matplotlib.text.Text:
                    annotations = label_annotations[event.artist.get_text()]
                    visible = annotations and not annotations[0].get_visible()
                    event.artist.set_alpha(visible_alpha[visible])
                    for annotation in annotations:
                        annotation.set_visible(visible)
            figure.canvas.mpl_connect('pick_event', onpick)
            adjust_toolbar(figure)
            figure.tight_layout()

    def plot_beam(self):
        with wait_cursor():
            with shut_up_mpl():
                figure, axis = plt.subplots()
            setup_number = self.beam.get_setup_index()
            wavelength = 300 / s.schsf.sffreq[0, setup_number-1]
            wave_text = "{} cm".format(
                "{:.1f}".format(wavelength * 100).rstrip("0").rstrip("."))
            source = self.beam.get_source()
            figure.canvas.set_window_title("Beam Plot ({}, {})".format(
                source, wave_text))

            aliases = set(next(s.aliases for s in self.sources 
                               if s.aliases[0] == source))
            
            def up_and_in(station, scan_index):
                return station.stascn[scan_index] and \
                    (station.up1[scan_index] == "") and \
                    (station.up2[scan_index] == "")

            # schedlib.plbeam expects 2 arrays for u and corresponding v points
            sched_uv_size = 600000 # defined in src/Plot/beam.inc
            u = np.zeros((sched_uv_size, 1))
            v = np.zeros((sched_uv_size, 1))
            uv_index = 0
            for scan_index, scan in enumerate(self.scan_catalog.used(), 
                                              self.scan_catalog.scan_offset):
                if not ((scan.scnsrc in aliases) and 
                        (scan.setnum == setup_number)):
                    continue
                station_uv = {station: np.array(s.stauv(scan_index+1, i+1)).\
                              reshape((2, 2)) / wavelength
                              for i, station in enumerate(self.stations)}
                baselines = itertools.combinations(self.stations, 2)
                for station1, station2 in baselines:
                    station_names = (station1.station, station2.station)
                    if self.beam.baselines.get_visible(station_names) and \
                       up_and_in(station1, scan_index) and \
                       up_and_in(station2, scan_index):
                        diff = station_uv[station1] - station_uv[station2]
                        u[uv_index:uv_index+2, 0] = diff[0, :]
                        v[uv_index:uv_index+2, 0] = diff[1, :]
                        uv_index += 2

            # multiply by factor, because of SCHED
            max_uv_dist = math.sqrt(np.max(u*u + v*v, axis=None)) * 10000 
            weighting_map = {"Natural": 0, "Uniform": 1}
            beam, xy_interval = s.plbeam(
                u, v, uv_index, max_uv_dist, 
                weighting_map[self.beam.get_weighting()],
                self.beam.get_oversampling())
            beam /= 1000
            itf = self.beam.get_image_transfer_function()
            if itf != "Linear":
                min_beam = np.min(beam, axis=None)
                max_beam = np.max(beam, axis=None)
                # normalize to [0, 1]
                beam = (beam - min_beam) / (max_beam - min_beam)
                if itf == "Logarithmic":
                    scale_factor = 65000 # copied from PGPLOT
                    beam = np.log(1 + scale_factor * beam) / \
                           math.log(1 + scale_factor)
                elif itf == "Square-root":
                    beam = np.sqrt((beam - min_beam) / (max_beam - min_beam))
                else:
                    raise RuntimeError("Unknown image transfer function {}".\
                                       format(itf))
                # scale back to [min, max]
                beam = beam * (max_beam - min_beam) + min_beam
            y, x = (np.mgrid[-beam.shape[1]//2:beam.shape[1]//2, 
                            -beam.shape[0]//2:beam.shape[0]//2] 
                    + 0.5) * xy_interval
            # transpose and flip beam to get the same picture as SCHED
            # (cannot find the documentation on meaning of the axes)
            beam = np.flip(beam.T, 0)

            image = axis.pcolormesh(x, y, beam,
                                    cmap=self.beam.get_color_map(),
                                    shading="auto")
            color_bar = figure.colorbar(image, ax=axis)
            if self.beam.get_contours():
                # levels copied from SCHED
                levels = [0.01, 0.03, 0.05, 0.08, 0.1, 0.3, 0.5, 0.8]
                min_beam = np.min(beam, axis=None)
                if min_beam < 0:
                    levels[0:4] = [min_beam * 2 / 3, min_beam / 3,
                                   -min_beam / 3, -min_beam * 2 / 3]
                    # plot negative contours with dashed lines
                    linestyles = ["dashed"] * 2 + ["solid"] * 6
                else:
                    linestyles = ["solid"] * 8
                overlay = axis.contour(x + 0.5 * xy_interval, 
                                       y + 0.5 * xy_interval, 
                                       beam, levels=sorted(set(levels)),
                                       linestyles=linestyles,
                                       cmap=self.beam.get_contour_colors())
                color_bar.add_lines(overlay)
                for line in color_bar.lines:
                    line.set_linewidth(5)
            
            axis.set_aspect("equal")
            axis.xaxis.set_label_text("RA (mas)")
            axis.invert_xaxis()
            axis.yaxis.set_label_text("Dec (mas)")
            axis.set_title("Beam for {} at {}".format(source, wave_text))
            adjust_toolbar(figure)
            figure.tight_layout()


def show(is_restart):
    """
    Return: (make files, do restart)
    """
    app = QApplication([])
    window = MainWidget(is_restart)
    window.setWindowTitle("SCHED Plot")
    window.finished.connect(app.closeAllWindows)
    window.show()
    plt.ion()
    app.exec_()
    r = window.result() # 0 -> exit, 1 -> finish, 2 -> restart
    return (r == 1, r == 2)



from .catalog import Catalog

import schedlib as s

class SourcePlotCatalog(Catalog):
    """
    CHARACTER    SRLNAME(MAXSRL)*12, SRLCALC(MAXSRL)*1
    CHARACTER    SRLEPO(MAXSRL)*1
    REAL         SRLRA(MAXSRL), SRLDEC(MAXSRL)
    REAL         SRLRAE(MAXSRL), SRLDECE(MAXSRL)
    LOGICAL      SRLUSED(MAXSRL)
    """

    max_source = s.srlis.srlused.shape[0]

    block_items = {
        s.srlis: [
            "srlra",
            "srlrae",
            "srldece",
            "srldec",
            "srlused"
        ],
        s.srcli: [
            "srlname",
            "srlcalc",
            "srlepo"
        ]
    }

    def __init__(self):
        super().__init__(self.max_source, self.block_items)

    def scheduled_slice(self):
        return slice(s.srlis.srln)
    



#!/usr/bin/env python3

from pysched.util import f2str
from pysched import key, update_catalogs
from pysched.version import check_version, pysched_version

from bottle import SimpleTemplate
import schedlib as s

import argparse
import sys
import io
import atexit
import os
import traceback

try:
    import readline
    histfile = os.path.join(os.path.expanduser("~"), ".pysched_history")
    try:
        readline.read_history_file(histfile)
    except FileNotFoundError:
        pass
    atexit.register(readline.write_history_file, histfile)
except ImportError:
    pass
except Exception as e:
    print("Warning, failed to import history file, error: {}".format(e))

parser = argparse.ArgumentParser(
    description=f"pySCHED version {pysched_version}, wrapper around SCHED")
parser.add_argument(
    "-t", "--template", nargs="?", const="", default=None, metavar="FILE",
    help="Input is a template which renders to KEYIN text using the following "
    "syntax: "
    "Expressions between {{...}} will be evaluated to strings. "
    "Python code lines start with %% and code blocks are surrounded by "
    "<%% and %%> tokens. All other text will be copied verbatim to the output. "
    "See http://bottlepy.org/docs/dev/stpl.html for the complete "
    "documentation. The rendered KEYIN text is then processed by "
    "SCHED. Since the template is rendered after all input is read, "
    "SCHED is non-interactive in this mode. If a parameter value is given, "
    "the resulting KEYIN text is written to that file. Warning: "
    "do not run an untrusted template, since it can contain arbitrary code.")
parser.add_argument("-p", "--plot", action="store_true", default=False,
                    help="Start plotting routines even if PLOT is not enabled "
                    "in the KEYIN input.")
parser.add_argument("-f", "--freqlist", nargs=2, type=float, default=None,
                    metavar=("lowF", "highF"),
                    help="Make frequency list (MHz). Then exit.")
parser.add_argument("-k", "--key", required=False, type=str,
                    help="Use the argument as input KEYIN schedule file.")
parser.add_argument("--old_style_vex", action="store_true", default=False,
                    help="Use the old VEX file printing function. "
                    "This function does not support all new features of "
                    "pySCHED.")
parser.add_argument("-U", "--no_update", action="store_true", default=False,
                    help="Do not automatically update the catalogs in {}.".\
                    format(update_catalogs.checkout_dir))
parser.add_argument("-v", "--version", action="store_true", default=False,
                    help="Print version and exit.")
parser.add_argument("-V", "--no_version_check", action="store_true",
                    default=False, help="Do not check for available updates "
                    "to pySCHED.")

args = parser.parse_args()

if args.version:
    print(pysched_version)
    sys.exit(0)

if not args.no_version_check:
    check_version()

try:
    if args.template is None:
        stdin = sys.stdin
    else:
        keyin_text = SimpleTemplate(sys.stdin.read()).render()
        stdin = io.StringIO(keyin_text)
        if args.template != "":
            with open(args.template, "w") as f:
                f.write(keyin_text)
            stdin.name = args.template
        else:
            stdin.name = "Rendered template"

    # point $SCHED to the checkout if not explicitly set, restore at exit
    if "SCHED" not in os.environ:
        os.environ["SCHED"] = update_catalogs.checkout_dir
        def restore():
            del os.environ["SCHED"]
        atexit.register(restore)

    # os.environ["SCHED"] is used in the initialization of sched modules,
    # so delay the import to here
    from pysched.sched import input_, parameter, schin_module, getfreq, \
        defaults, vexout, schopt, stmsg

    # workaround for https://github.com/numpy/numpy/issues/9370
    # summary of the different views of strings
    # python: strings are a type of their own, which includes an encoding
    # f2py: returns bytes when fortran would return character*, 
    #       but when assigning a bytes variable to a character*, 
    #       f2py will interpret it as a array of integers, convert each of those 
    #       integers to a string, and then trim that string to 1 character,
    #       for example b"az" -> [97, 122] -> ["97", "122"] -> "91".
    #       assigning a python string to an f2py generated fortran string gives
    #       the correct result
    # fortran: strings are fixed length character arrays, fortran doesn't have a
    #          end-of-string marker, it will just append blanks (spaces) in any
    #          string operation, eg "foo   " == "foo" evaluates to true.
    #          to mimic this behaviour, extend strings with spaces to fortran 
    #          length
    s.vern.vernum, version = s.verwrap()
    s.verc.version = bytes(version).decode().ljust(s.verc.version.itemsize)
    stmsg()

    if not args.no_update:
        update_catalogs.update_catalogs()

    if args.freqlist is not None:
        # first initialize default files stored in fortran common block 
        # state default values are (value, operator) tuples
        msgfile_default = schin_module.state_defaults["msgfile"]
        s.schsco.msgfile = msgfile_default[1](msgfile_default[0]).ljust(
            s.schsco.msgfile.itemsize)
        freqfile_default = schin_module.state_defaults["freqfile"]
        s.schsco.freqfile = freqfile_default[1](freqfile_default[0]).ljust(
            s.schsco.freqfile.itemsize)
        s.schcon.freqlist = args.freqlist
        getfreq()
        s.wlog(1, "DIVERT:   Frequency table written.  Stopping.")
        sys.exit(0)
    if args.key is not None:
        # set default schedule file
        schin_module.state_defaults["schedule"][0] = args.key

    restart = False

    while True:

        s.setn1.nset = 0
        s.srlis.srln = 0

        input_(stdin)
        print("SCHED:   Processing input.")
        defaults()
        s.schpre()
        s.chksc1()
        schopt()
        s.dopfq()
        s.getsun()
        s.chkscn()
        s.schsum(restart)
        s.fluxh(parameter.ilog, s.schsco.logfile)
        if s.schcon.plot or args.plot:
            # initializes matplotlib, so only do when requested
            from pysched import plot
            mkfiles, restart = plot.show(restart)
        else:
            mkfiles, restart = True, False

        if restart:
            s.delscr(restart)
            s.wlog(0, " ")
            s.wlog(0, " ===================  RESTART  ===================== ")
            s.wlog(0, " ")
        else:
            break

    if mkfiles and (f2str(s.schsco.optmode) != "UPTIME") and not s.schcon.noset:
        s.scnrange()
        s.omsout(restart)
        if args.old_style_vex:
            s.vexout()
        else:
            vexout()

        if s.schcon.dovsop:
            s.vsopwrt()

        s.flags()
        s.stafiles()

    s.delscr(False)
    s.putout(" -------  Finished  ----------- ")
except Exception as e:
    s.wlog(1, "Error running pySCHED: {e}".format(e=e))
    s.wlog(0, traceback.format_exc())



import schedlib as s

import numpy as np

def ifdbbc3():
    if s.schcon.debug:
        s.wlog(1, "IFDBBC3: Starting")
    max_bbc = 64
    max_if = 8
    ifbbc = np.zeros((max_bbc, max_if))
    for if_ in range(max_if):
        ifbbc[if_ * 8: (if_+1) * 8, if_] = 1
    return ifbbc, max_if



# BBC and IF assignments in the VEX file are currently ignored by eMERLIN 
# out stations, for historical reasons emulate a GEO DBBC assignment

from . import bbcalt, ifdbbc

import schedlib as s

import numpy as np

def bbc_emerlin(ks, setup_entry, station_entry):
    """
    ks: Fortran index into setups
    setup_entry: changes will be written to catalog
    station_entry: read only
    """
    if s.schcon.debug:
        s.wlog(1, "BBCEMERLIN: Starting")

    ifbbc, mif = ifdbbc("GEO", False) 
    ifnam = list("AB")
    max_bbc = ifbbc.shape[0]
    if station_entry.nbbc > max_bbc:
        s.wlog(1, "BBCDBBC: Number of VCs at {} Larger than maximum expected: "
             "{}".format(setup_entry.setsta[0], max_bbc))
        s.wlog(1, "   Catalog or programming problem ")
        s.errset(ks)
    ubbc = bbcalt(ks, setup_entry, ifbbc, ifnam, "eMERL", "BBCEMERLIN")



from ..catalog import ScanCatalog, SourceCatalog, SetupCatalog
from ..util import f2str, resize_string
from .. import vex

import schedlib as s

def vexout():
    if s.schcon.dovex:
        scan_catalog = ScanCatalog()
        scan_catalog.read()
        source_catalog = SourceCatalog()
        source_catalog.read()
        sources = source_catalog.scheduled()
        moving = False
        for scan in scan_catalog.used():
            source = sources[scan.srcnum - 1] # -1: adjust for FORTRAN indexing
            if source.sused and not scan.norec and \
               (source.planet or source.satel or 
                (source.dra >= 0.0001/15) or (source.ddec >= 0.0001)):
                moving = True
                break
        if moving:
            s.wlog(1, "++++ VEXOUT: Cannot pass accurate positions for planets "
                   "or satellites in the VEX file.")
            s.wlog(1, "        For correlation, get the source positions "
                   "elsewhere - like from ephemeris.")
            s.wlog(1, "        For VLBA pointing observations as of 2011, the "
                   "VEX positions")
            s.wlog(1, "        are not used so this is ok.  You may ignore the "
                   "warnings from VXSUDT.")

        setup_catalog = SetupCatalog()
        setup_catalog.read()
        setups = setup_catalog.used()
        all_none = all(setup.format == "NONE" for setup in setups)
        observation_type = f2str(s.schsco.obstyp)
        if (s.schcon.override or (not all_none) or \
            (observation_type == "PTVLBA")) and \
            (observation_type != "CONFIG"):
            # write two VEX (version) files, 
            # only print warnings while making the first one
            with open("{}.vex".format(f2str(s.schc1.expcode).lower()), "w") \
                 as vex_file:
                s.wlog(0, "Writing V E X file {}".format(vex_file.name))
                vex.write(vex_file, vex_version="1.5", print_warnings=True)
                # write VEX file name to common block for v2dout
                s.vex1.vexfile = resize_string(vex_file.name, 
                                               s.vex1.vexfile.itemsize,
                                               "vexfile")
            with open("{}.vex2".format(f2str(s.schc1.expcode).lower()), "w") \
                 as vex_file:
                s.wlog(0, "Writing V E X 2 file {}".format(vex_file.name))
                vex.write(vex_file, vex_version="2.0", print_warnings=False)

            if observation_type != "PTVLBA":
                s.v2dout()
        elif all_none:
            s.errlog("VEXOUT: Cannot use VEX with a schedule in which all "
                     "scans have FORMAT=NONE.")
        elif s.schn1.config:
            s.errlog("VEXOUT: Do not mix VEX with Configuration tests. "
                     "Set DOVEX=-1")



from .catalog import Catalog

import schedlib as s

class SetupCatalog(Catalog):
    """
    INTEGER           LISTKS(MSET)
    INTEGER           ISETNUM(MSET)
    INTEGER           IFREQNUM(MCHAN,MSET)
    INTEGER           IFREQIF(MCHAN,MSET)
    CHARACTER         SETNAME(MSET)*80
    INTEGER           SFCHAN(MCHAN,MSET), 
                      SGCHAN(MCHAN,MSET)
    DOUBLE PRECISION  FREQREF(MCHAN,MSET)
    DOUBLE PRECISION  FIRSTLO(MCHAN,MSET)
    DOUBLE PRECISION  FIFMIN(MCHAN,MSET)
    DOUBLE PRECISION  FIFMAX(MCHAN,MSET)
    DOUBLE PRECISION  BBSYN(MCHAN,MSET)
    DOUBLE PRECISION  BBSYN2(MCHAN,MSET)
    DOUBLE PRECISION  BBFILT(MCHAN,MSET)
    DOUBLE PRECISION  CORINV(MCHAN,MSET)
    DOUBLE PRECISION  SYNTH(3,MSET)
    REAL              AZCOLIM(MSET), 
                      ELCOLIM(MSET)
    REAL              PTINCR(MSET)
    REAL              PTOFF(MSET)
    REAL              SAMPRATE(MSET)
    REAL              SPEEDH(MSET)
    REAL              SPEEDL(MSET)
    REAL              FANOUT(MSET)
    REAL              SPEEDUP(MSET)
    REAL              BESTOVER(MSET)
    REAL              TOTBW(MSET)
    INTEGER           ISETSTA(MSET)
    INTEGER           NCHAN(MSET)
    INTEGER           TRACK(MCHAN,MTPMOD,MSET)
    INTEGER           PERIOD(MSET)
    INTEGER           BITS(MCHAN,MSET)
    INTEGER           LEVEL(MSET)
    INTEGER           BBC(MCHAN,MSET)
    INTEGER           TAPEMODE(MSET)      
    INTEGER           TPMODE(MSET)      
    INTEGER           SWTCHDUR(MSET)
    INTEGER           PCALFR1(MAXPC,MSET)
    INTEGER           PCALFR2(MAXPC,MSET)
    INTEGER           VFESYN(MCHAN,MSET)
    LOGICAL           DUALPOL(MSET)
    LOGICAL           FRSWITCH(MSET)
    LOGICAL           DUALX(MSET)
    LOGICAL           MODETEST(MSET)
    LOGICAL           USED(MSET)
    CHARACTER         SETSTA(MANT,MSET)*8
    CHARACTER         BAND(MSET)*5
    CHARACTER         IFDIST(4,MSET)*3
    CHARACTER         SIDEBD(MCHAN,MSET)*1
    CHARACTER         CRDSIDE(MCHAN,MSET)*1
    CHARACTER         POL(MCHAN,MSET)*4
    CHARACTER         NOISE(4,MSET)*6
    CHARACTER         NOISEFRQ(MSET)*4
    CHARACTER         FE(4,MSET)*6
    CHARACTER         IFCHAN(MCHAN,MSET)*2
    CHARACTER         ALTIFC(MCHAN,MSET)*2
    CHARACTER         LOGGING(MSET)*8
    CHARACTER         STRING(4,MSET)*80
    CHARACTER         FORMAT(MSET)*8
    CHARACTER         DBE(MSET)*8
    CHARACTER         FIRMFILE(MSET)*80
    CHARACTER         SPCAL(MSET)*4
    CHARACTER         PCALX1(MAXPC,MSET)*3
    CHARACTER         PCALX2(MAXPC,MSET)*3
    CHARACTER         NETSIDE(MCHAN,MSET)*1
    CHARACTER         SIDE1(MCHAN,MSET)*1
    CHARACTER         RCHAN(MSET)*2
    CHARACTER         LCHAN(MSET)*2
    CHARACTER         LCP50CM(MSET)*6
    CHARACTER         RCP50CM(MSET)*6
    CHARACTER         BARREL(MSET)*9
    CHARACTER         M4PATCH(MSET)*8
    INTEGER           MINTRAK(MSET)
    INTEGER           MAXTRAK(MSET)
    REAL              MINTBPS(MSET)
    REAL              MAXTBPS(MSET)
    REAL              TBPS(MSET)
    REAL              TOTBPS(MSET)
    REAL              WRTBPS(MSET)
    LOGICAL           VLBAMKIV(MSET)
    LOGICAL           RECUSED(MSET)
    INTEGER           FLUKESET(MSET)   
    DOUBLE PRECISION  FLUKEA(MSET)     
    DOUBLE PRECISION  FLUKEB(MSET)     
    DOUBLE PRECISION  VLAFEAB(MSET)    
    DOUBLE PRECISION  VLAFECD(MSET)    
    DOUBLE PRECISION  VLASYNA(MSET)    
    DOUBLE PRECISION  VLASYNB(MSET)    
    CHARACTER         FEFILTER(MSET)*4
    CHARACTER         VLAIF(MSET)*10
    CHARACTER         VLAROT(MSET)*10
    CHARACTER         VLABAND(MSET)*2
    CHARACTER         VLABW(MSET)*4
    LOGICAL           VLALOFI(MSET)
    LOGICAL           VLAVA(MSET)
    LOGICAL           VLAVB(MSET)
    LOGICAL           VLAVR(MSET)
    LOGICAL           VLAVL(MSET)
    """
    
    maxsetup = s.setn1.isetnum.shape[0]
    block_items = {
        s.setn1: [
            "listks",
            "isetnum",
            "ifreqnum",
            "ifreqif",
            "sfchan",
            "sgchan"
        ],
        s.setc1: [
            "setname"
        ],
        s.setn2a: [
            "freqref",
            "firstlo",
            "fifmin",
            "fifmax",
            "bbsyn",
            "bbsyn2",
            "bbfilt",
            "corinv",
            "synth",
            "azcolim",
            "elcolim",
            "ptincr",
            "ptoff",
            "samprate",
            "speedh",
            "speedl",
            "fanout",
            "speedup",
            "bestover",
            "totbw"
        ],
        s.setn2b: [
            "isetsta",
            "nchan",
            "track",
            "period",
            "bits",
            "level",
            "bbc",
            "tapemode",
            "tpmode",
            "swtchdur",
            "pcalfr1",
            "pcalfr2",
            "vfesyn",
            "dualpol",
            "frswitch",
            "dualx",
            "modetest",
            "used",
            "mintrak",
            "maxtrak",
            "mintbps",
            "maxtbps",
            "tbps",
            "totbps",
            "wrtbps",
            "vlbamkiv",
            "recused"
        ],
        s.setc2: [
            "setsta",
            "band",
            "ifdist",
            "sidebd",
            "crdside",
            "pol",
            "noise",
            "noisefrq",
            "fe",
            "ifchan",
            "altifc",
            "logging",
            "string_bn",
            "format",
            "dbe",
            "firmfile",
            "spcal",
            "pcalx1",
            "pcalx2",
            "netside",
            "side1",
            "rchan",
            "lchan",
            "lcp50cm",
            "rcp50cm",
            "barrel",
            "m4patch"
        ],
        s.setn3: [
            "flukeset",
            "flukea",
            "flukeb",
            "vlafeab",
            "vlafecd",
            "vlasyna",
            "vlasynb",
            "vlalofi",
            "vlava",
            "vlavb",
            "vlavr",
            "vlavl"
        ],
        s.setc3: [
            "fefilter",
            "vlaif",
            "vlarot",
            "vlaband",
            "vlabw",
        ]
    }

    def __init__(self):
        super().__init__(self.maxsetup, self.block_items, {"dbbcfw"})

    class Channel(object):
        attributes = {
            "ifreqnum",
            "ifreqif",
            "sfchan",
            "sgchan",
            "freqref",
            "firstlo",
            "fifmin",
            "fifmax",
            "bbsyn",
            "bbsyn2",
            "bbfilt",
            "corinv",
            "bits",
            "bbc",
            "vfesyn",
            "sidebd",
            "crdside",
            "pol",
            "ifchan",
            "altifc",
            "netside",
            "side1"}
        def __init__(self, entry, index):
            self.entry = entry
            self.index = index

        # forward attribute get and set to entry
        def __getattr__(self, attr):
            if attr in self.attributes:
                return getattr(self.entry, attr)[self.index]
            # track is the only multidimensional attribute, 
            # so do it separately
            elif attr == "track":
                return getattr(self.entry, attr)[:, self.index]
            raise AttributeError(attr)

        def __setattr__(self, attr, value):
            if attr in self.attributes:
                getattr(self.entry, attr)[self.index] = value
            # track is the only multidimensional attribute, 
            # so do it separately
            elif attr == "track":
                getattr(self.entry, attr)[:, self.index] = value
            else:
                super().__setattr__(attr, value)

    def read(self):
        ret = super().read()
        # nchan is not derived from the length of keyin channel parameters,
        # but is a parameter by itself. Therefore do the reduction of the 
        # size of arrays which depend on it here.
        for entry in ret:
            nchan = entry.nchan
            # track is the only multidimensional attribute, so do it separately
            entry.track = entry.track[:, :nchan]
            for attr in self.Channel.attributes:
                setattr(entry, attr, getattr(entry, attr)[:nchan])
            # create a view to all parameters accessed by channel index
            entry.channel = [self.Channel(entry, index) 
                             for index in range(nchan)]
        return ret
    
    def scheduled_slice(self):
        return slice(s.setn1.nset)

    def used(self):
        """
        Pre: catalog is read.
        """
        return [entry for entry in self.scheduled() if entry.used]
        
    @staticmethod
    def is_dbbc_e_firmware(entry):
        """
        Returns whether the setup entry represent a setup with e-series
        DBBC firmware (True|False).
        """
        if entry.dbbcfw == "":
            return False # assume 107 (or better)
        return entry.dbbcfw[-1] in ("E", "F")

    @classmethod
    def dbbc_firmware_allowed_sample_rates(cls, entry):
        """
        Returns the set of allowed sample rates for the configured 
        DBBC firmware.
        """
        if cls.is_dbbc_e_firmware(entry):
            return {4, 8, 16, 32, 64}
        if entry.dbbcfw == "": # assume 107 (or better)
            return {2, 4, 8, 16, 32, 64, 128}
        if entry.dbbcfw < "107": # known firmwares are 104 and 105
            return {2, 4, 8, 16, 32}
        return {2, 4, 8, 16, 32, 64}



from . import toggle
from .. import util

import schedlib as s

import numpy as np

def invla(values, present, entries, index, case):
    warn = True
    if s.schcon.debug and ((index < 2) or (case != 1)):
        s.wlog(0, "INVLA: Starting with case: {}".format(case))

    if case == 1:
        entry = entries[index]
        
        default = False
        if index > 0:
            default = (entries[index-1].vlatsys == "T")
        entry.vlatsys = "T" if toggle(values, present, "vlantsys", "vlatsys", 
                                      default) \
                        else " "

        if entry.vlaphs == "":
            entry.vlaphs = entry.scnsrc

        if warn and (("vlabw" in present) or ("vlaband" in present)):
            s.wlog(1, "INVLA: -- WARNING -- VLABW or VLABAND  specified in "
                   "main schedule.")
            s.wlog(1, "INVLA:               They will be ignored.")
            s.wlog(1, "INVLA:               Specify them in the setup files.")
            warn = False

    elif case == 2:
        s.schc3.vlatype = util.resize_string(values["vlatype"], 
                                             s.schc3.vlatype.itemsize, "vlatype")
        s.schn3.vlausern = values["vlausern"]
        s.schn1.iatutc = values["iatutc"]

        if not s.schcon.noset:
            if [station[:3] for station in s.schc1.staname[:s.schn1.nsta]].\
               count(b"VLA") >= 2:
                s.wrtmsg(1, "INVLA", "multipleVLA")
    else:
        s.errlog("INVLA: Bad case.")

    s.schn3.vlarfant = values["vlarfant"]



import requests
from distutils.version import LooseVersion

import pkg_resources

try:
    pysched_version = pkg_resources.get_distribution("pythonSCHED").version
except:
    pysched_version = "unknown"

def get_latest_version():
    response = requests.get("https://pypi.org/pypi/pythonSCHED/json")
    response.raise_for_status()
    return response.json()["info"]["version"]

def check_version():
    try:
        latest_version = get_latest_version()
        if LooseVersion(pysched_version) < LooseVersion(latest_version):
            print(f"Version {latest_version} of pySCHED is available through "
                  f"pip,\n you are running version {pysched_version}.")
    except Exception:
        pass



from ..catalog import SetupCatalog
from .. import util, key
from . import parameter

import schedlib as s

import itertools

dbbc_firmware_warnings = set()
def rdset(setreq, input_iterator, isetf):
    s.setn1.sdebug = s.schcon.debug
    if s.setn1.sdebug:
        print("RDSET SDEBUG=", s.setn1.sdebug)

    if s.schcon.debug:
        s.wlog(0, "RDSET: Reading setup {} {}".format(isetf, setreq))
    
    catalog = SetupCatalog()

    start = int(s.setn1.nset)
    index = start
    if next((entry for entry in catalog.entries[:index] \
             if entry.setname == setreq), 
            None) is not None:
        s.wlog(1, "RDSET: Duplicate setup file name? ")
        s.wlog(1, "RDSET:{}".format(setreq))
        s.errlog("Check setup file names")
    
    s.wlog(0, "RDSET:   Reading setup file:      {}".format(setreq))
        
    mchan = s.setn2a.freqref.shape[0]
    maxpc = s.setc2.pcalx1.shape[0]
    state_defaults = {
        "logging":  ["STANDARD",        util.noop],
        "noisefrq": ["VLBA",            util.upper],
        "fe":       [["omit"] * 4,      util.foreach(util.lower)],
        "noise":    [["low-s"] * 4,     util.noop],
        "ifdist":   [[0] * 4,           util.foreach(
            lambda x: x.lower() if isinstance(x, str) else str(int(x)))],
        "azcolim":  [0.,                util.noop],
        "elcolim":  [0.,                util.noop],
        "ptincr":   [0.,                util.noop],
        "ptoff":    [0.,                util.noop],
        "period":   [1,                 util.noop],
        "level":    [-1,                util.noop],
        "samprate": [0.,                util.noop],
        "dualx":    [1.,                util.to_bool],
        "synth":    [[0.] * 3,          util.noop],
        "frswitch": [1.,                util.to_bool],
        "tpmode":   [0.,                util.noop],
        "format":   ["",                util.upper],
        "modetest": [1.,                util.to_bool],
        "dbe":      ["",                util.upper],
        "swtchdur": [15.,               util.noop],
        "firmfile": ["",                util.noop],
        "barrel":   ["not_set",         util.lower],
        "tpspeedh": [0.,                util.noop],
        "tpspeedl": [0.,                util.noop],
        "tpspeed":  [0.,                util.noop],
        "lcp50cm":  ["DEF",             util.upper],
        "rcp50cm":  ["DEF",             util.upper],
        "rchan":    ["",                util.upper],
        "lchan":    ["",                util.upper],
        "nchan":    [0,                 int],
        "band":     ["",                util.lower],
        "firstlo":  [[parameter.unset], util.extend_to(mchan)],
        "bbsyn":    [[0.] * mchan,      util.noop],
        "bbsyn2":   [[0.] * mchan,      util.noop],
        "bbfilter": [[0.],              util.extend_to(mchan)],
        "sideband": [[" "],             util.chain(util.foreach(util.upper), 
                                                   util.extend_to(mchan))],
        "netside":  [[" "],             util.chain(util.foreach(util.upper), 
                                                   util.extend_to(mchan))],
        "ifchan":   [[""] * mchan,      util.foreach(util.upper)],
        "bits":     [[1],               util.extend_to(mchan)],
        "bbc":      [[0.] * mchan,      util.noop],
        "freqoff":  [[0.] * mchan,      util.noop],
        "freqref":  [[0.],              util.extend_to(mchan)],
        "pol":      [[""],              util.chain(util.foreach(util.upper), 
                                                   util.extend_to(mchan))],
        "track1":   [[0.] * mchan,      util.noop],
        "track2":   [[0.] * mchan,      util.noop],
        "track3":   [[0.] * mchan,      util.noop],
        "track4":   [[0.] * mchan,      util.noop],
        "track5":   [[0.] * mchan,      util.noop],
        "track6":   [[0.] * mchan,      util.noop],
        "track7":   [[0.] * mchan,      util.noop],
        "track8":   [[0.] * mchan,      util.noop],
        "string1":  ["",                util.noop],
        "string2":  ["",                util.noop],
        "string3":  ["",                util.noop],
        "string4":  ["",                util.noop],
        "pcal":     ["1MHZ",            util.upper],
        "pcalxb1":  [[""] * maxpc,      util.foreach(util.upper)],
        "pcalxb2":  [[""] * maxpc,      util.foreach(util.upper)],
        "pcalfr1":  [[0.] * maxpc,      util.noop],
        "pcalfr2":  [[0.] * maxpc,      util.noop],
        "m4patch":  ["astro",           util.upper],
        "dbbcfw":   ["",                util.chain(
            lambda x: str(x).rstrip("0").rstrip(".") 
            if type(x) == float else x, 
            util.upper)],
    }
    record_defaults = {
        "station": [[""] * s.setc2.setsta.shape[0], util.foreach(util.upper)],
        "endset":  [None,                           util.noop],
    }
    input_iterator.set_defaults(record_defaults, state_defaults)

    old_vla_parameters = {"flukeset", "flukea", "flukeb", "vlafeab", "vlafecd", 
                          "vlasyna", "vlasynb", "fefilter", "vlaif", "vlarot", 
                          "vlaband", "vlabw"}
    record_defaults.update(
        {key: [None, util.noop] for key in old_vla_parameters})

    attribute_to_key = {
        "setsta": "station",
        "speedh": "tpspeedh",
        "speedl": "tpspeedl",
        "sidebd": "sideband",
        "pcalx1": "pcalxb1",
        "pcalx2": "pcalxb2",
        "bbfilt": "bbfilter",
        "spcal": "pcal",
    }
    # update with default identity mapping, 
    # skip keys not stored directly
    attribute_names = set(itertools.chain(state_defaults.keys(), 
                                          record_defaults.keys())) - \
        {"tpspeed", 
         "freqoff",
         "track1",
         "track2",
         "track3",
         "track4",
         "track5",
         "track6",
         "track7",
         "track8",
         "string1",
         "string2",
         "string3",
         "string4",
         "endset",
        } - \
        old_vla_parameters
    attribute_to_key.update({attribute: attribute 
                             for attribute in attribute_names
                             if attribute not in attribute_to_key.values()})

    for record in input_iterator:
        values, present = util.merge_record_with_defaults(
            record, record_defaults, state_defaults)
        if "endset" in present:
            break

        if index >= catalog.maxsetup:
            s.errlog("RDSET: Too many setup groups ")
        
        entry = catalog.entries[index]
        entry.set_keyin_values(values, attribute_to_key)

        entry.setname = setreq
        entry.isetnum = isetf
        if entry.ptoff <= 0.:
            entry.ptoff = 6. * entry.ptincr
        if (entry.speedl == 0.) and (values["tpspeed"] != 0.):
            entry.speedl = values["tpspeed"]
        if entry.nchan > mchan:
            s.errlog("RDSET: Too many  channels in {}".format(setreq))
        
        justvla = all(station[:3] in ("VLA", "") for station in entry.setsta) \
                  and (entry.format == "NONE") and (entry.nchan == 0)
        if not justvla:
            if entry.nchan < 1:
                s.errlog("RDSET: NCHAN must be greater than 0.")

            entry.freqref = [entry.freqref[i] + values["freqoff"][i]
                             for i in range(entry.nchan)]

            for ichan in range(entry.nchan):
                pol_map = {"R": "RCP",
                           "L": "LCP"}
                if entry.pol[ichan] in pol_map.keys():
                    entry.pol[ichan] = pol_map[entry.pol[ichan]]
                elif (entry.ifchan[ichan] in pol_map.keys()) and \
                     (entry.pol[ichan] == ""):
                    entry.pol[ichan] = pol_map[entry.ifchan[ichan]]
                    
                if entry.ifchan[ichan] == "R":
                    entry.ifchan[ichan] = entry.rchan
                if entry.ifchan[ichan] == "L":
                    entry.ifchan[ichan] = entry.lchan

            entry.track = [values["track" + n] for n in "12345678"]
        #enf of if not justvla
        
        entry.string_bn = [values["string" + n] for n in "1234"]
        if any(values[key] is not None for key in old_vla_parameters):
            s.wlog(1, "RDSET - WARNING: Old VLA parameters used in input "
                   "setup: {}".format(index + 1))
            s.wlog(1, "                 They are no longer used and will be "
                   "removed soon. ")

        pcal_map = {
            "OFF": "off",
            "1MHZ": "1MHz",
            "5MHZ": "5MHz"
        }
        if entry.spcal in pcal_map.keys():
            entry.spcal = pcal_map[entry.spcal]
        else:
            s.errlog("CHKSET: Invalid PCAL specification: {}".format(
                entry.spcal))

        if (entry.dbbcfw not in ("", "104", "105", "105E", "105F", "106E", 
                                 "106F", "107")) and \
            (entry.dbbcfw not in dbbc_firmware_warnings):
            s.wlog(1, "RDSET - WARNING: {} is not a known DBBC firmware "
                   "version.".format(entry.dbbcfw))
            s.wlog(1, "                 Heuristics might still work, but check "
                   "schedule carefully.")
            dbbc_firmware_warnings.add(entry.dbbcfw)
        
        index += 1

    s.setn1.nset = index
    catalog.write(range(start, index))



from ..catalog import LineRestFrequencyCatalog
from .. import util, key

import schedlib as s

s.schlin.nlgp = 0

def rfreq(input_iterator):
    if s.schcon.debug:
        s.wlog(0, "RFREQ:  Starting to read line frequencies")
    maxlch = s.schlin.restfreq.shape[0]

    record_defaults = {
        "lineset":  ["",   util.upper],
        "restfreq": [[0.], util.extend_to(maxlch)],
        "endlines": [None, util.noop],
    }
    input_iterator.set_defaults(record_defaults, {})

    attribute_to_key = {
        "linename": "lineset",
        "restfreq": "restfreq",
    }
    catalog = LineRestFrequencyCatalog()
    start = int(s.schlin.nlgp)
    index = start

    found_end = False
    for record in input_iterator:
        values, present = util.merge_record_with_defaults(
            record, record_defaults, {})
        if "endlines" in present:
            found_end = True
            break
        if index >= catalog.maxlgp:
            s.errlog(" SCHED only set up to use {} rest frequency groups.".\
                     format(catalog.maxlgp))
        entry = catalog.entries[index]
        entry.set_keyin_values(values, attribute_to_key)
        index += 1

    if not found_end:
        s.errlog("RFREQ: Input data ended while reading spectral line rest "
                 "frequencies.")

    catalog.write(range(start, index))

    s.schlin.nlgp = index



import schedlib as s

def check_emerlin(setup, station):
    bits = max(setup.bits)
    expected_samprate = 256 // bits
    if setup.samprate != expected_samprate:
        s.wlog(1, "Check eMERLIN: Invalid SAMPRATE specified for station {}: "
               "{} for DAR=eMERL. Must be {} Msamp/s for {} bits.".format(
                   station.station, setup.samprate, expected_samprate, bits))
        return True
    return False



from . import srread, rdpeak, stread, rfreq, satini, pcread, rdset
from ..catalog import SetupFileCatalog

import schedlib as s

def schfiles(input_iterator, stdin, values, present, gotsat):
    restart = False
    if s.schcon.plot and (input_iterator.input_ is stdin):
        s.errlog("SCHFILES: Must read input from SCHEDULE file when plotting.")

    if "tapeini" in present:
        s.errlog("TAPEINI sections no longer supported in SCHED")
        
    if "satinit" in present:
        if gotsat:
            s.errlog("SCHFILES: Can only read one SATINI section.")
        gotsat = True

    def handle_setinit():
        setinit = values["setinit"]
        index = SetupFileCatalog.extend_with(setinit, 
                                             "SCHFILES: Too many setup files.")
        
        rdset(setinit, input_iterator, index)
            
    # handle inline catalogs, possible ones are:
    inline_catalogs = [
        ("srccat", "Imbedded source catalog.", 
         lambda: srread(input_iterator, stdin, False, 'i')),
        ("peakinit", "Imbedded peak inst.", 
         lambda: rdpeak(input_iterator, stdin)),
        ("stacat", "Imbedded station catalog.", 
         lambda: stread(input_iterator, stdin, mjd1=0.)), # mjd1 is unused
        ("lineinit", "Reading line freqs.", 
         lambda: rfreq(input_iterator)),
        ("satinit", "Reading satellite info.", 
         lambda: satini(input_iterator)),
        ("setinit", "Imbedded setup: {}".format(values["setinit"]),
         handle_setinit),
        ("pcenters", "Phase center lists.", 
         lambda: pcread(input_iterator))
    ]
    todo = [cat for cat in inline_catalogs if cat[0] in present]
    # can do only one at a time
    if len(todo) > 1:
        s.errlog("'SCHFILES: Cannot do 2 in-stream files at the same time.")
    for keyword, text, function in todo:
        if s.schcon.debug:
            s.wlog(1, "SCHFILES: {}".format(text))
        function()
        restart = True

    if "coverlet" in present:
        # copy text verbatim to temporary file 
        # until "endcover" and "/" are on one line
        try:
            temp_file = open("schedcov.tmp", "w")
        except Exception as e:
            s.wlog(1, str(e))
            s.errlog("SCHFILES: Cannot open cover letter scratch file.  "
                     "This should not happen.")
        with temp_file:
            while True:
                line = input_iterator.parser.readline()
                if len(line) == 0:
                    s.errlog("SCHFILES: Cover letter not terminated with "
                             "ENDCOVER /")
                if (line.upper().find("ENDCOVER") != -1) and \
                   (line.find("/") != -1):
                    break
                temp_file.write(line)
        s.schcon.coverlet = True
        restart = True
        
    return gotsat, restart



import os

# git import will raise an exception if it cannot find the executable,
# this can be suppressed with the environment variable GIT_PYTHON_REFRESH
_var = "GIT_PYTHON_REFRESH"
_old_value = os.environ.get(_var)
os.environ[_var] = "quiet"
import git
if _old_value is not None:
    os.environ[_var] = _old_value

import schedlib as s

import sys

checkout_dir = os.path.join(os.path.expanduser("~"), ".pysched")
git_repository = "https://github.com/jive-vlbi/sched.git"
branch = "data_files_v1.4"

class Spinner(git.RemoteProgress):
    cursor_chars = ["|", "/", "-", "\\"]
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        sys.stdout.write(" ")
        sys.stdout.flush()
        self.index = 0
    
    def update(self, op_code, cur_count, max_count=None, message=""):
        sys.stdout.write("\r")
        sys.stdout.write(self.cursor_chars[self.index])
        sys.stdout.flush()
        self.index = (self.index + 1) % len(self.cursor_chars)

def update_catalogs():
    try:
        if not os.path.exists(checkout_dir):
            s.wlog(1, "Downloading catalogs to {}".format(checkout_dir))
            # clone a shallow, single branch version to reduce data usage
            repo = git.Repo.clone_from(git_repository, checkout_dir, 
                                       depth=1, branch=branch, 
                                       progress=Spinner())
            print()
        else:
            repo = git.Repo(checkout_dir)

        # make sure the required branch is in the local repo
        if branch not in {b.name for b in repo.branches}:
            repo.git.remote("set-branches", "origin", branch)
            repo.git.fetch("origin", branch, depth=1)

        repo.git.checkout(branch)

        s.wlog(1, "Updating catalogs in {}".format(checkout_dir))
        repo.git.pull()
        s.wlog(1, "Catalogs in {} are up-to-date.".format(checkout_dir))

    except git.exc.GitError as e:
        s.wlog(1, "Warning, failed to update catalogs.")
        s.wlog(0, "Update failed with error: {}".format(e))



import schedlib as s

import numpy as np
from .. import util

def get_arrays(attr_array):
    # gather a copy (because .T returns a view) of all arrays in C-order
    vector_func = np.vectorize(util.f2str, otypes=[object])
    arrays = {attr: vector_func(array.T) if array.dtype.kind == "S" 
              else array.T.copy()
              for attr, array in attr_array.items()}
    return arrays

def write(attr_array, entries, indices):
    for attr, array in attr_array.items():
        dtype = array.dtype
        vector_func = np.vectorize(
            lambda x: util.resize_string(x, dtype.itemsize, attr), 
            otypes=[dtype])
        for i in indices:
            new_value = np.array(getattr(entries[i], attr), dtype=dtype)
            if dtype.kind == "S":
                # fill string with spaces fortran/sched style
                new_value = vector_func(new_value)
            # if the new value is smaller, embed it
            dest_slice = (slice(i, i+1), ) + \
                         tuple(map(slice, new_value.shape))
            array.T[dest_slice] = new_value

def map_attr_array(block_items):
    return {attr: getattr(block, attr)
            for block, attrs in block_items.items()
            for attr in attrs}

class Catalog(object):
    """
    Base class to help read f2py common blocks in python objects
    and write back.
    Instances of subclasses share the state with each other.
    """

    # FIX: replace indexing member to other catalog entries 
    #      with actual catalog entry

    class CatalogEntry(object):
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)
        
        def set_keyin_values(self, values, attribute_to_key):
            for attribute, key in attribute_to_key.items():
                if key in values:
                    setattr(self, attribute, values[key])

        def __str__(self):
            return str(self.__dict__)
           

    def __init__(self, nr_elements, block_items, extended_attributes=set()):
        """
        nr_elements: the maximum number of catalog entries
        block_items: {schedlib block: [attribute of block]}
        extended_attributes: {name of attribute not represented in 
                              Fortran common blocks}
        """
        try:
            self.__dict__ = type(self)._shared_state
            assert(self.nr_elements == nr_elements)
            assert(self.block_items == block_items)
            assert(self.extended_attributes == extended_attributes)
        except AttributeError:
            # no shared state for this class yet, create and assign
            self.__dict__ = type(self)._shared_state = {}
            self.nr_elements = nr_elements
            self.block_items = block_items
            self.attr_array = map_attr_array(block_items)
            self.attributes = list(self.attr_array.keys())
            self.extended_attributes = extended_attributes
            self.prime()

    def adjust_lengths(self, entries):
        """
        Where the common blocks contain fixed length arrays,
        set the length of the entries using the variable defining the length.
        """
        pass
    
    def prime(self):
        """
        Prepares the catalog with the values from common blocks to overwrite
        with keyin values before calling self.write().
        """
        arrays = get_arrays(self.attr_array)
        # create an entry for each index of the arrays
        self.entries = [
            self.CatalogEntry(
                **{key: value[i] for key, value in arrays.items()})
            for i in range(self.nr_elements)]
        for entry in self.entries:
            for attribute in self.extended_attributes:
                setattr(entry, attribute, None)
        self.adjust_lengths(self.entries)
        return self.entries

    def read(self, selection_slice=None):
        """
        Read the catalog entries from the common blocks.
        Only read and return entries that sliced by selection_slice, 
        applied to self.entries.
        If selection_slice is None, self.scheduled_slice() is used.
        """
        if selection_slice is None:
            selection_slice = self.scheduled_slice()
        arrays = get_arrays(self.attr_array)
        entries = self.entries[selection_slice]
        for i, entry in zip(range(*selection_slice.indices(len(self.entries))), 
                            entries):
            entry.__dict__.update(
                {key: value[i] for key, value in arrays.items()})
        self.adjust_lengths(entries)
        return entries

    def scheduled_slice(self):
        """
        Return the slice of entries that are scheduled.
        """
        return slice(None)
    
    def scheduled(self):
        """
        Return entries that have defined values.
        """
        return self.entries[self.scheduled_slice()]

    def write(self, indices=None):
        """
        Write the entries to common blocks.
        Where the common blocks contain fixed length arrays,
        set the variable defining the length from 
        the length of the entries (the inverse process of prime). 
        
        indices: write only these entries to the catalog.
        """
        if indices is None:
            indices = range(self.nr_elements)
        write(self.attr_array, self.entries, indices)



from .. import util
from . import parameter
from ..update_catalogs import checkout_dir
from ..version import pysched_version

import schedlib as s

import os.path

def stmsg():
    s.schcon.readlog = False
    logfile = "sched.runlog"
    s.schsco.logfile = util.resize_string(logfile, s.schsco.logfile.itemsize,
                                          "logfile")
    open_stat = "OLD" if os.path.exists(logfile) else "NEW"
    error_msg, error_code = s.openwrap(parameter.ilog, logfile, "TEXT",
                                       open_stat)
    if error_code != 1:
        s.wlog(1, f"STMSG: CANNOT OPEN LOG FILE {logfile}")
        s.wlog(1, f"       Error text: {error_msg}")
        s.error(" Check why log file cannot be opened")

    s.wlog(1, " ")
    s.wlog(1, f"    Welcome to pySCHED version {pysched_version} "
           "based on ")
    s.wlog(1, f"    SCHED version: {s.vern.vernum:.6} "
           f"{util.f2str(s.verc.version)}")
    s.wlog(1, " ")
    s.wlog(1, "The manual is at http://www.aoc.nrao.edu/software/sched/"
           "index.html")
    s.wlog(1, f"pySCHED will use the catalogs under {checkout_dir} unless "
           "$SCHED is set.")
    s.wlog(1, f"Most run time messages will be in {logfile}")

    s.schsco.freqver = util.resize_string("None used",
                                          s.schsco.freqver.itemsize, "freqver")

    s.wlog(1, " ")
    s.wlog(1, f"Check 'sched.py -h' for command line paramters.")
    s.wlog(1, "Some useful commands to give now if running interactively:")
    s.wlog(1, "   SCHedule=<filename>    :  Specify input file.")
    s.wlog(1, "   PLOT                   :  Invokes uv, xy, rd, and uptime "
           "plotting.")
    s.wlog(1, "   FREQLIST=lowF,highF    :  Make frequency list (MHz).  "
           "Then exit.")
    s.wlog(1, "   EXIT                   :  Leave program.")
    s.wlog(1, "   /                      :  End of inputs - run program "
           "(or EXIT).")
    s.wlog(1, " ")
    
    if s.rdcatn.srcvel.shape[0] < s.schn2b.freq.shape[0]:
        s.errlog("STMSG: Programming error.  MVEL .LT. MAXCHN")

    



from . import bbcalt, ifdbbc, ifdbbc3
from ..catalog import SetupCatalog

import schedlib as s

import numpy as np

def bbcdbbc(ks, setup_entry, station_entry):
    """
    ks: Fortran index into setups
    setup_entry: changes will be written to catalog
    station_entry: read only
    """
    if s.schcon.debug:
        s.wlog(1, "BBCDBBC: Starting")

    if station_entry.dar == "DBBC3":
        ifbbc, mif = ifdbbc3() 
        ifnam = list("ABCDEFGH")
    else:
        e_firmware = SetupCatalog.is_dbbc_e_firmware(setup_entry)
        ifbbc, mif = ifdbbc(station_entry.dbbcver, e_firmware) 
        ifnam = list("ABCD")
    max_bbc = ifbbc.shape[0]
    if station_entry.nbbc > max_bbc:
        s.wlog(1, "BBCDBBC: Number of VCs at {} Larger than maximum expected: "
             "{}".format(setup_entry.setsta[0], max_bbc))
        s.wlog(1, "   Catalog or programming problem ")
        s.errset(ks)
    ubbc = bbcalt(ks, setup_entry, ifbbc, ifnam, "DBBC", "BBCDBBC")



from . import gmkscn
from .parameter import max_seg, secpday
from ..catalog import StationCatalog
from ..util import f2str, bool2str

import schedlib as s

import numpy as np

import math

def makeseg(j_scan, scan_index, last_scan_index, ok_geo, use_geo, seg_elevation, 
            start_time, end_time, dummy_index, sigma, stations, scans):
    
    if s.schcon.debug:
        s.wlog(1, "MAKESEG starting")

    n_sta = len(stations)
    n_print = min(20, n_sta)

    max_geo = s.schsou.geosrci.shape[0]
    selected = np.zeros(dtype=int, shape=(max_geo,))

    t_src = np.zeros(dtype=int, shape=(max_seg,))
    selection_type = np.full(fill_value="", dtype=object, shape=(max_seg,))
    worst_station = np.full(fill_value="", dtype=object, shape=(max_seg,))
    in_scan = np.zeros(dtype=int, shape=(n_sta,))
    station_low = np.zeros(dtype=int, shape=(n_sta,))
    station_high = np.zeros(dtype=int, shape=(n_sta,))
    n_station_low = 0
    n_station_high = 0
    low_high_goal = 1
    last_high_station = 0
    last_high_sigma = 0

    # otherwise used uninitialized in log output ...
    high_station = 0
    high_sigma = 0
    save_sigma1 = np.copy(sigma)
    save_sigma2 = np.copy(sigma)
    save_sigma3 = np.copy(sigma)
    high_station1 = 0
    high_station2 = 0
    high_station3 = 0

    last_scan_index_copy = np.copy(last_scan_index)

    scans[scan_index - 1].startj = start_time

    if s.schsou.geoprt >= 1:
        s.wlog(0, "".join("{:^6}".format(s.stcode) for s in stations[:n_print]))

    seg_index = 0
    do_rand = True
    nt_seg = 0
    max_prio = 6

    scan_stascn = np.empty(dtype=bool, shape=(n_sta,))
    k_stascn = np.empty(dtype=bool, shape=(max_geo, n_sta))

    chance = np.empty(dtype=int, shape=(max_seg * 10,))
    test_time = np.empty(dtype=float, shape=(max_seg * 10,))

    while True:
        seg_index += 1

        if seg_index > max_seg:
            s.wlog(1, " ")
            s.wlog(1, " Trying to exceed SCHED''s limit of {} scans in an "
                   "inserted geodetic segment.".format(max_seg))
            s.errlog(" Shorten the requested segment or divide it into "
                     "multiple segments.")

        l_scan = scan_index + seg_index - 1

        scan_range = slice(scan_index - 1, l_scan - 1)
        
        # scans in segment per station where station is up
        n_stascn = np.array([np.count_nonzero(
            np.logical_and(s.stascn[scan_range] != 0,
                           np.logical_and(f2str(s.up1[scan_range]) == "",
                                          f2str(s.up2[scan_range]) == "")))
                             for s in stations], dtype=int)

        if s.schsou.geoprt >= 2:
            s.wlog(0, " ")
            s.wlog(0, "----MAKESEG:  Starting to work on next source: "
                   "ISEG: {}   DORAND: {}   MINNSS: {}   MAXPRIO: {}".format(
                       seg_index, bool2str(do_rand), min(n_stascn), max_prio))
        
        ii_scan = max(scan_index, l_scan - s.schsou.geoback + 1)
        s_scan1  = max(scan_index, l_scan - s.schsou.geosrep + 1)

        if (((n_station_low < n_sta) or (n_station_high < n_sta)) and do_rand) \
           or (min(n_stascn) <= 1):
            method = "RAND"

            n_near = 5
            n_chance = 0

            for geo_index in range(1, s.schsou.ngeo + 1):
                if (use_geo[geo_index - 1] <= max_prio) and \
                   (selected[geo_index - 1] < s_scan1):
                    if seg_index == 1:
                        approx_time = start_time
                    else:
                        approx_time = scans[l_scan - 2].stopj + 60 / secpday

                    if s.schsou.geoprt >= 2:
                        s.wlog(0, "MAKESEG random section - looking for nearby "
                               "source.  Considering: {}  SRC: {} {}  "
                               "USEGEO: {} at {:7.6f}".format(
                                   seg_index, geo_index, 
                                   f2str(s.schcsc.geosrc[geo_index - 1]), 
                                   use_geo[geo_index - 1], approx_time))

                    n_good, ok_sta, scan_stascn = gmkscn(
                        last_scan_index_copy, l_scan, j_scan, 
                        s.schsou.geosrci[geo_index - 1], 
                        f2str(s.schcsc.geosrc[geo_index - 1]), approx_time, 
                        scans[j_scan - 1].opminel, 0, scan_stascn, "SET")

                    k_stascn[geo_index - 1, :] = scan_stascn

                    n_new_low = 0
                    n_new_high = 0
                    need_low = 0
                    need_high = 0
                    for i, station in enumerate(stations):
                        if station_low[i] < low_high_goal:
                            need_low += 1
                        if station_high[i] < low_high_goal:
                            need_high += 1
                        if scan_stascn[i]:
                            average_el = (station.el1[l_scan - 1] + 
                                          station.el2[l_scan - 1]) / 2
                            if (average_el <= s.schsou.geolowel) and \
                               (station_low[i] < low_high_goal):
                                n_new_low += 1
                            if (average_el >= s.schsou.geohiel) and \
                               (station_high[i] < low_high_goal):
                                n_new_high += 1
                                
                    src_ok = (n_new_low + n_new_high > n_sta // 3) or \
                             (n_new_low + n_new_high >= 
                              max(1, (need_low + need_high) // 3))

                    sta_ok = True
                    new_c = False
                    for i, station in enumerate(stations):
                        temp_in_scan = in_scan[i]
                        if station.stascn[l_scan - 1]:
                            temp_in_scan += 1
                        if temp_in_scan < seg_index // 2:
                            sta_ok = False

                    if src_ok and sta_ok and \
                       (n_good >= scans[j_scan - 1].opmian):
                        if (n_chance < n_near) or (l_scan == s.schn1.scan1):
                            n_chance += 1
                            chance[n_chance - 1] = geo_index
                            test_time[n_chance - 1] = scans[l_scan - 1].startj
                            new_c = True
                        elif scans[l_scan - 1].startj < test_time[n_near - 1]:
                            chance[n_near - 1] = geo_index
                            test_time[n_near - 1] = scans[l_scan - 1].startj
                            new_c = True
                            
                        if s.schsou.geoprt >= 2:
                            if l_scan == s.schn1.scan1:
                                l_source_name = "First scan"
                                seperation_time = 0
                            else:
                                l_source_name = scans[l_scan - 2].scnsrc
                                seperation_time = \
                                    (scans[l_scan - 1].startj - 
                                     scans[l_scan - 2].stopj) * secpday
                            
                            s.wlog(0, "MAKESEG getting slew info {} {} "
                                   "Last source: {}  Testing: {}".format(
                                       geo_index, l_scan, l_source_name, 
                                       f2str(s.schcsc.geosrc[geo_index - 1])))

                            for i, station in enumerate(stations, 1):
                                if l_scan == s.schn1.scan1:
                                    l_stas = False
                                else:
                                    l_stas = station.stascn[l_scan - 2]

                                lsi = last_scan_index_copy[i - 1] - 1
                                if lsi < 0:
                                    el2 = 0
                                    az2 = 0
                                else:
                                    el2 = station.el2[lsi]
                                    az2 = station.az2[lsi]

                                s.wlog(0, " makeseg els, azs, tslew: {:4d} {:4d}"
                                       " {:7.1f} {:7.1f} {:7.1f} {:7.1f} {:7.1f}"
                                       " {} {} {:7.1f}".format(
                                           i, lsi + 1, 
                                           el2, station.el1[l_scan - 1], 
                                           az2, station.az1[l_scan - 1], 
                                           station.tslew[l_scan - 1] * secpday, 
                                           bool2str(l_stas), 
                                           bool2str(station.stascn[l_scan - 1]), 
                                           seperation_time))
                                
                        if new_c and (l_scan != s.schn1.scan1):
                            for ic in range(n_chance - 1, 0, -1):
                                if test_time[ic] < test_time[ic - 1]:
                                    # swap
                                    chance[ic - 1], chance[ic] = \
                                        chance[ic], chance[ic - 1]
                                    test_time[ic - 1], test_time[ic] = \
                                        test_time[ic], test_time[ic - 1]
                        
                    #if src_ok and sta_ok and (n_good >= scan[j_scan - 1].opmian)
                #if (use_geo[igeo] <= max_prio) and (selected[igeo] < s_scan1)
            #for igeo in range(1, ngeo + 1)
            
            if n_chance == 0:
                if s.schsou.geoprt >= 1:
                    s.wlog(0, "MAKESEG: No priority <= {} sources found for "
                           "random source additions.".format(max_prio))

                if 3 <= max_prio < 5:
                    if min(n_stascn) >= 2:
                        do_rand = False
                        if s.schsou.geoprt >= 1:
                            s.wlog(0, "         Select a source using a "
                                   "SecZ fit.")
                    else:
                        max_prio += 1
                        do_rand = True
                elif max_prio >= 5:
                    s.wlog(1, "Too few useful sources to choose from to "
                           "construct a geodetic segment. ")
                    s.wlog(1, "   Possible solutions include: ")
                    s.wlog(1, "     Add sources ")
                    s.wlog(1, "     Adjust GEOLOWEL and GEOHIEL so more "
                           "sources are considered useful.")
                    s.wlog(1, "     Try different OPMINANT or OPMINEL")
                    s.wlog(1, "    Pick a solution and try again.")
                else:
                    max_prio += 1
                    if s.schsou.geoprt >= 1:
                        s.wlog(0, "         Allow scans with low stations but "
                               "no very high elevation stations.")
                
                seg_index -= 1
                continue

            max_ic = 0
            for ic in range(1, n_chance + 1):
                if test_time[ic - 1] <= test_time[0] + 60 / secpday:
                    max_ic = ic
            
            dummy_index, result = s.ran5wrap(dummy_index)
            ic = int(1 + result * max_ic)
            
            if s.schsou.geoprt >= 2:
                s.wlog(0, "MAKESEG picked random source: {} {}".format(
                    ic, chance[ic - 1]))

            t_src[seg_index - 1] = chance[ic - 1]

            scan_stascn[:] = k_stascn[t_src[seg_index - 1] - 1, :]
                
            selection_type[seg_index - 1] = "NR"
            worst_station[seg_index - 1] = "--"

            if s.schsou.geoprt >= 2:
                t0 = int(scans[l_scan - 1].stopj)
                minpday = 24 * 60
                minutes = (scans[l_scan - 2].stopj - t0) * minpday
                s.wlog(0, "MAKESEQ slew times: {} {} {:.2f} {} {}".format(
                    t_src[seg_index - 1], ic, minutes, n_chance, max_ic))
                s.wlog(0, " Time below is minutes in day ")
                s.wlog(0, "                    Src    Time  Since Last stop")
                for is_ in range(n_chance):
                    s.wlog(0, "{:9d} {:9d} {:6} {:10.2f} {:9.2f}".format(
                        is_ + 1, chance[is_], 
                        f2str(s.schcsc.geosrc[chance[is_] - 1]),
                        (test_time[is_] - t0) * minpday, 
                        (test_time[is_] - scans[l_scan - 2].stopj) * minpday))

        #if (((n_station_low < n_sta) or (n_station_high < n_sta)) and do_rand) \
        #   or (min(n_stascn) <= 1):
        else:
            method = "FIT"
            bsl_qual = 1e9
            bf_qual = 1e9
            bf_qual_sl = 1e9
            t_src[seg_index - 1] = 0
            t_src1 = 0
            t_src2 = 0
            t_src3 = 0
            for geo_index in range(1, s.schsou.ngeo + 1):
                if ok_geo[geo_index - 1] and (selected[geo_index - 1] < s_scan1):
                    approx_time = scans[l_scan - 2].stopj + 60 / secpday
                    n_good, ok_sta, scan_stascn = gmkscn(
                        last_scan_index_copy, l_scan, j_scan, 
                        s.schsou.geosrci[geo_index - 1], 
                        f2str(s.schcsc.geosrc[geo_index - 1]), approx_time, 
                        scans[j_scan - 1].opminel, last_high_station,
                        scan_stascn, "SET")
                    
                    k_stascn[geo_index - 1, :] = scan_stascn

                    sta_ok = True

                    if last_high_station != 0:
                        if not scan_stascn[last_high_station - 1]:
                            sta_ok = False

                    if sta_ok:
                        for i, station in enumerate(stations):
                            temp_in_scan = in_scan[i]
                            if station.stascn[l_scan - 1]:
                                temp_in_scan += 1
                            if temp_in_scan < seg_index // 2:
                                sta_ok = False

                            if not station.stascn[l_scan - 1]:
                                stations_in_lookback = np.count_nonzero(
                                    station.stascn[ii_scan - 1: l_scan - 1])
                                if stations_in_lookback < 2:
                                    sta_ok = False
                    
                    if (n_good >= scans[j_scan - 1].opmian) and sta_ok and \
                       (scans[l_scan - 1].stopj <= end_time):
                        if t_src[seg_index - 1] == 0:
                            k_scan = ii_scan
                        else:
                            k_scan = l_scan

                        print_debug = (s.schsou.geoprt >= 2)
                        if print_debug:
                            s.wlog(0, " ")
                            s.wlog(0, "Makeseg about to test quality {} {} {} "
                                   "{} {} {}  geosrc: {} {}".format(
                                       seg_index, scan_index, ii_scan, k_scan, 
                                       l_scan, j_scan, geo_index, 
                                       f2str(s.schcsc.geosrc[geo_index - 1])))
                        
                        test_quality, sigma = s.geoqual(
                            ii_scan, k_scan, l_scan, j_scan, 14.48, print_debug, 
                            sigma)

                        if last_high_station >= 1:
                            f_qual = sigma[last_high_station - 1]
                        else:
                            f_qual = 0

                        sl_qual = math.sqrt(np.sum(np.square(sigma[:n_sta])) 
                                            / n_sta)
                        max_sigma = np.max(sigma[:n_sta])

                        slew_time = scans[l_scan - 1].startj - \
                                    scans[l_scan - 2].stopj
                        penalty = s.schsou.geoslew * slew_time / (30 / secpday)

                        sl_qual += penalty
                        if f_qual != 0:
                            f_qual_sl = f_qual + penalty
                        else:
                            f_qual_sl = 0

                        if s.schsou.geoprt >= 2:
                            s.wlog(0, "Makeseg quality - LSCN: {} {}  Source: "
                                   "{} {}  Quality: {:.2f} {:.2f} {:.2f} {} "
                                   "{:.2f}".format(
                                       l_scan, seg_index, geo_index,  
                                       f2str(s.schcsc.geosrc[geo_index - 1]), 
                                       f_qual_sl, f_qual, sl_qual, 
                                       last_high_station, max_sigma))

                        if (f_qual_sl != 0) and (f_qual_sl < bf_qual_sl):
                            bf_qual_sl = f_qual_sl
                            t_src1 = geo_index
                            high_station1 = np.argmax(sigma[:n_sta]) + 1
                            high_sigma1 = sigma[high_station1 - 1]
                            save_sigma1 = np.copy(sigma)
                            if s.schsou.geoprt >= 2:
                                s.wlog(0, "makeseg: Got new best sigma with "
                                       "slew penalty:  {}  {} {:.2f} {} {:.2f}".\
                                       format(
                                           geo_index, 
                                           f2str(s.schcsc.geosrc[geo_index - 1]),
                                           f_qual_sl, high_station1, 
                                           high_sigma1))
                            
                        if (f_qual != 0) and (f_qual < bf_qual):
                            bf_qual = f_qual
                            t_src2 = geo_index
                            high_station2 = np.argmax(sigma[:n_sta]) + 1
                            high_sigma2 = sigma[high_station2 - 1]
                            save_sigma2 = np.copy(sigma)
                            if s.schsou.geoprt >= 2:
                                s.wlog(0, "makeseg: Got new best sigma without "
                                       "slew penalty:  {}  {} {:.2f} {} {:.2f}".\
                                       format(
                                           geo_index, 
                                           f2str(s.schcsc.geosrc[geo_index - 1]),
                                           f_qual, high_station2, 
                                           high_sigma2))

                        if sl_qual < bsl_qual:
                            bsl_qual = sl_qual
                            t_src3 = geo_index
                            high_station3 = np.argmax(sigma[:n_sta]) + 1
                            high_sigma3 = sigma[high_station3 - 1]
                            save_sigma3 = np.copy(sigma)
                            if s.schsou.geoprt >= 2:
                                s.wlog(0, "makeseg: Got new best RMS sigma with "
                                       "slew penalty:  {}  {} {:.2f} {} {:.2f}".\
                                       format(
                                           geo_index, 
                                           f2str(s.schcsc.geosrc[geo_index - 1]),
                                           sl_qual, high_station3, 
                                           high_sigma3))

                    #if (n_good >= scans[j_scan - 1]) and sta_ok and \
                    #   (scans[l_scan - 1] <= end_time)
                #if ok_geo[geo_index - 1] and (selected[geo_index - 1] < s_scan1)
            #for geo_index in range(1, s.schsou.ngeo + 1):
            
            if last_high_station != 0:
                if (t_src1 != 0) and \
                   ((last_high_sigma - save_sigma1[last_high_station - 1]) > 
                    (last_high_sigma - save_sigma2[last_high_station - 1]) 
                    * 0.7) and \
                   (last_high_sigma / save_sigma1[last_high_station - 1] > 1.1):
                    t_src[seg_index - 1] = t_src1
                    selection_type[seg_index - 1] = "SP"
                    worst_station[seg_index - 1] = \
                        stations[high_station1 - 1].stcode

                    high_station = high_station1
                    high_sigma = high_sigma1
                    save_sigma = np.copy(save_sigma1)
                    last_type = "SIGP"
                    if s.schsou.geoprt >= 1:
                        s.wlog(0, "MAKESEG: {}  Choice used slew penalty. "
                               "Last: {:.2f}  Options (sig/sta/src): "
                               "{:.2f}/{}/{} {:.2f}/{}/{} {:.2f}/{}/{}".format(
                                   seg_index, last_high_sigma, 
                                   save_sigma1[last_high_station - 1], 
                                   high_station1, t_src1, 
                                   save_sigma2[last_high_station - 1], 
                                   high_station2, t_src2, 
                                   save_sigma3[last_high_station - 1], 
                                   high_station3, t_src3))
                    
                elif (t_src2 != 0) and \
                   ((last_high_sigma / save_sigma2[last_high_station - 1] > 1.25)
                    or ((last_type == "RMS") and 
                        (high_station3 == last_high_station) and
                        (high_sigma3 > high_sigma2))):
                    t_src[seg_index - 1] = t_src2
                    selection_type[seg_index - 1] = "S"
                    worst_station[seg_index - 1] = \
                        stations[high_station2 - 1].stcode
                    
                    high_station = high_station2
                    high_sigma = high_sigma2
                    save_sigma = np.copy(save_sigma2)
                    last_type = "SIG"
                    if s.schsou.geoprt >= 1:
                        s.wlog(0, "MAKESEG: {}  Choice used sigma only. "
                               "Last: {:.2f}  Options (sig/sta/src): "
                               "{:.2f}/{}/{} {:.2f}/{}/{} {:.2f}/{}/{}".format(
                                   seg_index, last_high_sigma, 
                                   save_sigma1[last_high_station - 1], 
                                   high_station1, t_src1, 
                                   save_sigma2[last_high_station - 1], 
                                   high_station2, t_src2, 
                                   save_sigma3[last_high_station - 1], 
                                   high_station3, t_src3))
            # if last_high_station != 0
    
            if (t_src3 != 0) and (t_src[seg_index - 1] == 0):
                t_src[seg_index - 1] = t_src3
                selection_type[seg_index - 1] = "RMS"
                worst_station[seg_index - 1] = stations[high_station3 - 1].stcode
                
                high_station = high_station3
                high_sigma = high_sigma3
                save_sigma = np.copy(save_sigma3)
                last_type = "RMS"
                if s.schsou.geoprt >= 1:
                    s.wlog(0, "MAKESEG: {}  Choice used RMS sigma.    "
                           "Last: {:.2f}  Options (sig/sta/src): "
                           "{:.2f}/{}/{} {:.2f}/{}/{} {:.2f}/{}/{}".format(
                               seg_index, last_high_sigma,
                               save_sigma1[last_high_station - 1], 
                               high_station1, t_src1, 
                               save_sigma2[last_high_station - 1], 
                               high_station2, t_src2, 
                               save_sigma3[last_high_station - 1], 
                               high_station3, t_src3))
                
            if t_src[seg_index - 1] != 0:
                scan_stascn[:] = k_stascn[t_src[seg_index - 1] - 1, :]
        
        #end of else 
        #if (((n_station_low < n_sta) or (n_station_high < n_sta)) and do_rand)
        #   or (min(n_stascn) <= 1):
        
        if t_src[seg_index - 1] > 0:
            if s.schsou.geoprt >= 2:
                s.wlog(0, "Makeseg got the next source - seg: {}  "
                       "Geosrc: {}  {} tapprox: {:.2f}  stascn: {}".format(
                           seg_index, t_src[seg_index - 1], 
                           f2str(s.schcsc.geosrc[t_src[seg_index - 1] - 1]), 
                           approx_time, 
                           " ".join(bool2str(e) for e in scan_stascn[:n_sta])))

            n_good, ok_sta, scan_stascn = gmkscn(
                last_scan_index_copy, l_scan, j_scan, 
                s.schsou.geosrci[t_src[seg_index - 1] - 1], 
                f2str(s.schcsc.geosrc[t_src[seg_index - 1] - 1]), approx_time, 
                scans[j_scan - 1].opminel, 0, scan_stascn, "FORCE")
            
            for i, station in enumerate(stations):
                if station.stascn[l_scan - 1]:
                    in_scan[i] += 1
                    last_scan_index_copy[i] = l_scan
                    average_el = (station.el1[l_scan - 1] + 
                                  station.el2[l_scan - 1]) / 2
                    if average_el <= s.schsou.geolowel:
                        station_low[i] += 1
                    if average_el >= s.schsou.geohiel:
                        station_high[i] += 1
                        
            n_station_low = np.count_nonzero(station_low >= low_high_goal)
            n_station_high = np.count_nonzero(station_high >= low_high_goal)
            
            selected[t_src[seg_index - 1] - 1] = l_scan

            if s.schsou.geoprt >= 1:
                msg = "Makeseg source: {} {}  {}".format(
                    seg_index, t_src[seg_index - 1], 
                    f2str(s.schcsc.geosrc[t_src[seg_index - 1] - 1]))
                if method == "FIT":
                    msg += " High: {}  Sigmas: {}".format(
                        high_station, 
                        " ".join("{:5.1f}".format(e) 
                                 for e in save_sigma[:n_print]))
                s.wlog(0, msg)

            last_high_station = high_station
            last_high_sigma = high_sigma

            if scans[l_scan - 1].stopj > \
               start_time + 0.7 * (end_time - start_time):
                do_rand = False
            else:
                do_rand = True

            max_prio = 2
            continue

        else:
            if s.schsou.geoprt >= 2:
                s.wlog(0, "makeseg:  No source found that fits")
                s.wlog(0, "          Ending sequence.")

            seg_index -= 1
            nt_seg = seg_index
            l_scan = scan_index + nt_seg - 1
            
            # this is the only exit point (apart from s.errlog calls)

            return l_scan, nt_seg, t_src, dummy_index, selection_type, \
                worst_station



import schedlib as s

import numpy as np

def ifdbbc(dbbc_version, e_firmware):
    if s.schcon.debug:
        s.wlog(1, "IFDBBC: Starting")
    max_bbc = 16
    max_if = 4
    ifbbc = np.zeros((max_bbc, max_if))
    if dbbc_version == "ASTRO":
        mif = 4
        for if_ in range(mif):
            ifbbc[if_ * 4: (if_+1) *4, if_] = 1
    elif dbbc_version == "GEO":
        mif = 2
        for if_ in range(mif):
            ifbbc[if_ * 8: (if_+1) *8, if_] = 1
    elif dbbc_version == "HYBRID":
        mif = 3
        ifbbc[0:4, 0] = 1
        ifbbc[4:8, 1] = 1
        ifbbc[8:16, 2] = 1
    else:
        s.wlog(1, "IFDBBC: DBBCVER not recognised: {}".format(dbbc_version))
        s.errlog("catalog error")

    if e_firmware: 
        # 32 MHz (only in E) DDC mode can only do odd (count starting at 1) BBCs
        ifbbc[1::2,:] = 0
    return ifbbc, mif



from ..catalog import PeakCatalog, SetupFileCatalog
from . import rfreq, schdefs
from .. import util, key

import schedlib as s

# two versions of the actual function, one that opens a file 
# and one that just uses the current keyin iterator

def rdpeak_open(stdin):
    if (s.schcon.debug):
        s.wlog(0, "RDPEAK starting")

    s.schpeakn.dopoint = s.schcon.autopeak or \
                         (s.schn2a.point[:s.schn1.nscans] >= 0).any()
    peakfile = bytes(s.schsco.peakfile).decode().strip()
    if (peakfile.upper() == "NONE") or not s.schpeakn.dopoint:
        return
    try:
        f = open(peakfile, "r")
    except Exception as e:
        s.wlog(1, str(e))
        s.wlog(1, "RDPEAK: Automatic insertion and/or conversion of scans "
               "was requested.")
        s.wlog(1, "        PEAKFILE needed, but could not be opened.")
        s.wlog(1, "        Problem is with file: ")
        s.wlog(1, peakfile)
        s.wlog(1, "        Note that file peak.cmd could be used")
        s.wlog(1, "        It is with the standard catalogs.")
        s.errlog(" Fix PEAKFILE or do not invoke AUTOPEAK or POINT.")
    with f:
        rdpeak_implementation(key.KeyfileIterator(f), stdin)

def rdpeak(input_iterator, stdin):
    if (s.schcon.debug):
        s.wlog(0, "RDPEAK starting")

    s.schpeakn.dopoint = True
    rdpeak_implementation(input_iterator, stdin)

first_call = True
def rdpeak_implementation(input_iterator, stdin):
    global first_call
    if first_call:
        s.schpeakn.pkgroup.fill(0)
        s.schpeakn.npkgrp = 0

    first_call = False

    input_name = "Program input" if input_iterator.input_ is stdin \
                 else input_iterator.input_.name
    
    state_defaults = {
        "srcfile":  [schdefs("refpointing"),          
                           util.noop],
        "setup":    ["",   util.expand_file_name],
        "setupl":   ["",   util.expand_file_name],
        "minfreq":  [60e3, util.noop],
        "minel":    [30.,  util.noop],
        "dwell":    [60.,  util.multiply_by(1 / 86400.)],
        "linename": ["",   util.upper],
        "vlamode":  ["",   util.upper],
    }
    record_defaults = {
        "stations": [[],   util.foreach(util.upper)],
        "sources":  [[],   util.foreach(util.upper)],
        "lineinit": [None, util.noop],
        "endpeak":  [None, util.noop],
    }
    input_iterator.set_defaults(record_defaults, state_defaults)
    
    # map from catalog entry attributes to keyin keywords
    attribute_to_key = {
        "pklines": "linename",
        "pvlamode": "vlamode",
        "pkdwell": "dwell",
        "pkminfq": "minfreq",
        "pkminel": "minel",
        "psetfile": "setup",
        "plsetfil": "setupl",
        "pksta": "stations",
        "pksrc": "sources",
    }

    catalog = PeakCatalog()
    start = int(s.schpeakn.npkgrp)
    index = start
    for record in input_iterator:
        values, present = util.merge_record_with_defaults(
            record, record_defaults, state_defaults)
        if "endpeak" in present:
            break
        if "lineinit" in present:
            rfreq(input_iterator)
            continue

        if index >= catalog.maxpeak:
            s.errlog("RDPEAK:  Too many reference pointing groups (Max {}) "
                     "in: {}".format(catalog.maxpeak, input_name))

        entry = catalog.entries[index]
        entry.set_keyin_values(values, attribute_to_key)

        if entry.plsetfil == "":
            entry.plsetfil = entry.psetfile
            s.wlog(1, "RDPEAK:  WARNING:  Your peak command file does not "
                   "have a separate setup file")
            s.wlog(1, "         for spectral line (narrow band) sources.  "
                   "You may encounter errors" )
            s.wlog(1, "         because the digital backends require "
                   "sample rate = 2 times bandwidth ")
            s.wlog(1, "         and the sample rate can only be changed "
                   "with a new setup file.")

        def check_pointing_setup(filename):
            return SetupFileCatalog.extend_with(
                filename, 
                "RDPEAK: Exceeded limit on number of setup files while adding "
                "ones needed for reference pointing.")
        
        entry.pklset = check_pointing_setup(entry.psetfile)
        entry.pklsetl = check_pointing_setup(entry.plsetfil)
        if (len(entry.pksta) == 0) and index > 0:
            entry.pksta = catalog.entries[index-1].pksta
        elif len(entry.pksta) == 0:
            s.errlog("RDPEAK:  First peak group has no stations.")

        if (len(entry.pksrc) == 0) and index > 0:
            entry.pksrc = catalog.entries[index-1].pksrc
        elif len(entry.pksrc) == 0:
            s.errlog("RDPEAK:  First peak group has no sources.")

        index += 1

    if index == 0:
        s.wlog(0, "RDPEAK:  No pointing instruction groups found in ")
        s.wlog(0, input_name)

    s.schpeakn.npkgrp = index

    s.schpeakc.psrcfile = util.resize_string(util.expand_file_name(
        state_defaults["srcfile"][0]), s.schpeakc.psrcfile.itemsize, "srcfile")

    catalog.write(range(start, index))




from . import rdset
from ..catalog import SetupFileCatalog, SetupCatalog
from .. import key

import schedlib as s

def getset():
    if s.schcon.debug:
        s.wlog(0, "GETSET: Starting.")

    for index, entry in enumerate(SetupFileCatalog().read()):
        setfilename = entry.setfile
        if setfilename in {"DUMMY", "DEFAULT"}:
            s.errlog(" SETUP file required. ")
        else:
            if setfilename not in \
               (entry.setname for entry in SetupCatalog().read()):
                try:
                    f = open(setfilename, "r")
                except Exception as e:
                    s.wlog(1, str(e))
                    s.errlog("RDSET: Problem opening setup file")
                with f:
                    input_iterator = key.KeyfileLister(f)
                    rdset(setfilename, input_iterator, index+1)

    s.setn2a.fifmin[:, :s.setn1.nset].setfield(0., dtype=s.setn2a.fifmin.dtype)
    s.setn2a.fifmax[:, :s.setn1.nset].setfield(1e15, 
                                                dtype=s.setn2a.fifmax.dtype)
    s.setn2a.corinv[:, :s.setn1.nset].setfield(0., dtype=s.setn2a.corinv.dtype)



from .catalog import Catalog

import schedlib as s

class FrequencyCatalog(Catalog):
    """
    INTEGER           PRIO(MFREQ)
    INTEGER           FNIF(MFREQ)
    CHARACTER         FSTNAM(MFSTA,MFREQ)*8
    CHARACTER         FIFNAM(MFIF,MFREQ)*2
    CHARACTER         FALTIF(MFIF,MFREQ)*2
    DOUBLE PRECISION  FRF1(MFIF,MFREQ), FRF2(MFIF,MFREQ)
    DOUBLE PRECISION  FLO1(MFIF,MFREQ)
    CHARACTER         FRNAME(MFREQ)*12
    CHARACTER         FRNOTE(MFREQ)*80
    CHARACTER         FPOL(MFIF,MFREQ)*3
    CHARACTER         FFE(MFIF,MFREQ)*4
    DOUBLE PRECISION  FSYN(3,MFREQ)
    DOUBLE PRECISION  FCH1RF1(MFIF,MFREQ), FCH1RF2(MFIF,MFREQ)
    CHARACTER         FRCP50CM(MFREQ)*6, FLCP50CM(MFREQ)*6
    LOGICAL           FDUALX(MFREQ)
    CHARACTER         FVCHNSTA(MFIF,MFREQ)*8
    CHARACTER         FVBAND(MFREQ)*2
    CHARACTER         FVBW(MFREQ)*4
    CHARACTER         FVFILT(MFREQ)*4
    DOUBLE PRECISION  FVFEAB(MFREQ), FVFECD(MFREQ)
    DOUBLE PRECISION  FVSYNA(MFREQ), FVSYNB(MFREQ)
    DOUBLE PRECISION  FVFLKA(MFREQ), FVFLKB(MFREQ)
    """

    maxfreq = s.frqn.prio.shape[0]
    block_items = {
        s.frqn: [
            "prio",
            "fnif",
            "frf1",
            "frf2",
            "flo1",
            "fsyn",
            "fch1rf1",
            "fch1rf2",
            "fdualx",
            "fvfeab",
            "fvfecd",
            "fvsyna",
            "fvsynb",
            "fvflka",
            "fvflkb"],
        s.frqc: [
            "fstnam",
            "fifnam",
            "faltif",
            "frname",
            "frnote",
            "fpol",
            "ffe",
            "frcp50cm",
            "flcp50cm",
            "fvchnsta",
            "fvband",
            "fvbw",
            "fvfilt"]
    }
    
    def __init__(self):
        super().__init__(self.maxfreq, self.block_items, {"tscal"})

    def write(self, indices=None):
        if indices is None:
            entries = self.entries
        else:
            entries = [self.entries[i] for i in indices]
        for entry in entries:
            entry.fnif = len(entry.fifnam)
        return super().write(indices)

    def adjust_lengths(self, entries):
        for entry in entries:
            length = entry.fnif
            for attribute in ("fifnam", "faltif", "frf1", "frf2", "flo1",
                              "fpol", "ffe", "fch1rf1", "fch1rf2", "fvchnsta"):
                setattr(entry, attribute, getattr(entry, attribute)[:length])
    
    def scheduled_slice(self):
        return slice(s.frqn.nfreq)



from ..catalog import SetupCatalog, StationCatalog

import schedlib as s

import numpy as np

import copy

def setexpnd():
    if s.schcon.debug:
        s.wlog(0, "SETEXPND: Starting.")

    for i in range(s.setn1.nset):
        s.setstdef(i+1)

    catalog = SetupCatalog()
    catalog.read()
    setup_entries = catalog.scheduled()
    copy_to = len(setup_entries)
    for setup_entry in setup_entries:
        for station in setup_entry.setsta[setup_entry.setsta != ""]:
            if copy_to >= len(catalog.entries):
                s.wlog(1, "SETEXPND: There are too many setup groups after "
                       "creating new ones")
                s.wlog(1, "          for each station in the input setups.")
                s.wlog(1, "          Maximum is {} which is setups times "
                       "stations.".format(len(catalog.entries)))
                s.errlog("SETEXPND: You need fewer setups or a bigger MSET in "
                         "SCHED.")
            dest = catalog.entries[copy_to]
            dest.__dict__.update(copy.deepcopy(setup_entry.__dict__))
            dest.setsta[0] = station
            copy_to += 1

    if copy_to == len(setup_entries):
        s.errlog("SETEXPND: None of the specified setups are used.")

    # move the new entries to the front and invalidate the others
    catalog.entries = catalog.entries[len(setup_entries):] + \
                      catalog.entries[:len(setup_entries)]
    s.setn1.nset = copy_to - len(setup_entries)

    station_catalog = StationCatalog()
    station_catalog.read()
    stations = station_catalog.used()
    for ks, setup_entry in enumerate(catalog.scheduled(), 1):
        setup_entry.listks = ks
        station_index = next(s.schn1.stanum[station.ischsta-1] 
                             for station in stations
                             if station.station == setup_entry.setsta[0])
        if station_index is None:
            s.wlog(1, "SETEXPND: Did not get station number. "
                   "Programming problem.")
            s.wlog(1, "          Station: {}".format(setup_entry.setsta[0]))
            s.errset(ks)
        setup_entry.isetsta = station_index

    catalog.write(range(s.setn1.nset))

    for lsta in range(s.schn1.nsta):
        for iscn in range(s.schn1.nscans):
            s.schn2a.nsetup[iscn, lsta] = s.gnset(iscn+1, lsta+1)





from .. import util

import schedlib as s

def getcov(values):
    s.schco.schver = values["version"]
    for attribute in [
            "piname",
            "phone",
            "email",
            "fax",
            "obsphone",
            "obsmode"]:
        setattr(s.schsco, attribute, 
                util.resize_string(values[attribute], 
                                   getattr(s.schsco, attribute).itemsize, 
                                   attribute))
    for i in range(4):
        address = "address" + str(i+1)
        s.schsco.address[i] = util.resize_string(
            values[address], s.schsco.address.itemsize, address)
        note = "note" + str(i+1)
        s.schsco.note[i] = util.resize_string(
            values[note], s.schsco.note.itemsize, note)


    if s.schcon.debug:
        s.wlog(0, "GETCOV: Checking cover information.")

    missing = False
    if s.schco.schver == 0.:
        s.wlog(1, "     Schedule version is missing. ")
        missing = True
    if values["piname"] == "":
        s.wlog(1, "     No PINAME given. ")
        missing = True
    if values["address1"] == "":
        s.wlog(1, "     No address specified.")
        missing = True
    if values["phone"] == "":
        s.wlog(1, "     No PI phone number specified.")
        missing = True
    if (values["email"] == "") and (values["fax"] == ""):
        s.wlog(1, "     No email address or fax number specified.")
        missing = True

    if missing:
        s.wlog(1, "GETCOV:  Cover information incomplete or missing.")
        if not s.schn1.notape and not s.schcon.plot:
            s.errlog("GETCOV: Cover information is required for VLBI "
                     "observations.")
        if s.schcon.plot:
            s.wlog(1, "GETCOV: Sched will plot, but not write telescope "
                   "control files.")
    
    text = [
        "Schedule Version: {:10.2f}".format(float(s.schco.schver)),
        "Processed by SCHED version: {:6.2f}  {}".format(
            float(s.vern.vernum), bytes(s.verc.version).decode()),
        "PI:       {}".format(values["piname"]),
        "Address:  {}".format(values["address1"]),
        "          {}".format(values["address2"]),
        "          {}".format(values["address3"]),
        "          {}".format(values["address4"]),
        "Phone:    {}".format(values["phone"]),
        "EMAIL:    {}".format(values["email"]),
        "Fax:      {}".format(values["fax"]),
        "Phone during observation: {}".format(values["obsphone"]),
        "Observing mode: {}".format(values["obsmode"]),
        "Notes:    {}".format(values["note1"]),
        "          {}".format(values["note2"]),
        "          {}".format(values["note3"]),
        "          {}".format(values["note4"])
    ]
    line = next((line for line in text if line.find("!") != -1), None)
    if line is not None:
        s.wlog(1, "GETCOV: Please do not use exclamation marks in the cover "
               "information.")
        s.wlog(1, "        They mess up the parsing of the CRD files at the "
               "VLBA stations.")
        s.wlog(1, "        One was found in the line: ")
        s.wlog(1, line)
        s.errlog("SCHIN: Remove exclamation marks.")

    for index, line in enumerate(text):
        s.schsco.cover[index] = util.resize_string(
            line, s.schsco.cover.itemsize, "cover")




from . import setexpnd, setdefs, chkset

import schedlib as s

def defset():
    setexpnd()
    setdefs()
    s.chkfirm()
    for i in range(s.setn1.nset):
        chkset(i + 1)
        
    s.chksfil()
    s.sfinfo()
    s.sbpair()
    s.chkvdifx()
    s.chkjive()



from .. import util

import schedlib as s

import numpy as np

def gintent(intents, entries, scan_index):
    entry = entries[scan_index]
    if len(intents) > 0:
        entry.nscint = len(intents)
        for intent_index, intent in enumerate(intents):
            # find in stored intents
            padded = util.resize_string(intent, s.schc2c.intent.itemsize,
                                        "intents")
            stored_indices = np.argwhere(
                s.schc2c.intent[:s.schn2a.nintent] == padded.encode())
            if len(stored_indices) == 0:
                if s.schn2a.nintent >= len(s.schc2c.intent):
                    s.errlog("GINTENT: Too many distinct intents. "
                             "Max {}  Report need for more.".format(
                                 len(s.schc2c.intent)))
                s.schc2c.intent[s.schn2a.nintent] = padded
                s.schn2a.nintent += 1
                entry.iscint[intent_index] = s.schn2a.nintent
            else:
                entry.iscint[intent_index] = stored_indices[0] + 1

        if "NONE" in (intent.upper() for intent in intents):
            if len(intents) > 1:
                s.errlog("GINTENT:  Please do not mix 'NONE' with "
                         "other INTENTs")
            entry.iscint[0] = 0
            entry.nscint = 0

    else:
        if scan_index + 1 > s.schn1.scan1:
            prev = entries[scan_index-1]
            entry.nscint, entry.iscint = prev.nscint, prev.iscint



from . import bbcdbbc, bbc_emerlin

import schedlib as s

# SCHED's version only has the ks parameter, 
# add the catalogs to prevent reloading
def setbbc(ks, setup_catalog, frequency_entries, station_entries):
    if s.setn1.sdebug:
        s.wlog(0, "SETBBC: Starting.")

    setup_entry = setup_catalog.entries[ks-1]
    for ich, channel in enumerate(setup_entry.channel):
        if channel.ifreqnum >= 1:
            if channel.ifchan == "":
                channel.ifchan = frequency_entries[channel.ifreqnum-1].\
                                 fifnam[channel.ifreqif-1]
        else:
            if channel.ifchan == "":
                s.wlog(1, "SETBBC: Cannot set IFCHANs.  First bad channel: "
                       "{} of {} total.".format(ich + 1, 
                                                len(setup_entry.channel)))
                s.errset(ks)
    setup_catalog.write(range(ks-1, ks))
    
    station_entry = station_entries[setup_entry.isetsta-1]
    if (station_entry.recorder == "S2") and \
       (station_entry.dar in ("VLBA", "VLBAG", "VLBA4")):
        s.bbcvs2(ks)
    else:
        def handle_func(f):
            def handle(ks):
                f(ks, setup_entry, station_entry)
                setup_catalog.write(range(ks-1, ks))
            return handle
        bbc_func = {"VLBA": s.bbcvlba,
                    "RDBE": s.bbcrdbe,
                    "RDBE2": s.bbcrdbe,
                    "WIDAR": s.bbcwidar,
                    "VLBAG": s.bbcgeo,
                    "VLBA4": s.bbcgeo,
                    "MKIV": s.bbcm4,
                    "DBBC": handle_func(bbcdbbc),
                    "DBBC3": handle_func(bbcdbbc),
                    "CDAS": s.bbccdas,
                    "R1002": s.bbckvsr,
                    "LBA": s.bbclba,
                    "eMERL": handle_func(bbc_emerlin)}.get(
                        station_entry.dar, None)
        if bbc_func is not None:
            bbc_func(ks)
        elif station_entry.dar != "NONE":
            s.wlog(1, "SETBBC: SCHED does not set default BBCs for format: {}, "
                   "DAR type: {}".format(setup_entry.format, station_entry.dar))
            s.errset(ks)
